# -*- coding: utf-8 -*-
#
# Copyright (c) 2006 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""provide view classes used to generate Documentor/Docbook views of the project
"""
from projman import format_monetary
from projman.lib.constants import HOURS_PER_DAY
from projman.lib._exceptions import ViewException
try:
    import xml.etree.ElementTree as ET
except ImportError:
    import elementtree.ElementTree as ET
    
# dom utilities ################################################################

DR_NS = "{http://www.logilab.org/2004/Documentor}"
LDG_NS = "{http://www.logilab.org/2005/DocGenerator}"
ET._namespace_map[DR_NS[1:-1]] = "dr"
ET._namespace_map[LDG_NS[1:-1]] = "ldg"

PUCE = u'\u203A '

def document(root=None):
    """return a DOM document node"""
    root = ET.Element(DR_NS + (root or "root") )
    return ET.ElementTree(root)

class DocbookHelper:
    """a helper class to generate docboock"""

    def __init__(self, lang='fr'):
        self.lang = lang

    def object_node(self, parent, task_id):
        """create a DOM node <section> with a attribute id"""
        assert isinstance(task_id, basestring)
        node = ET.SubElement( parent, DR_NS + "object", id=task_id, lang=self.lang )
        return node

    def table_layout_node(self, parent, nbcols, align='left', colsep=1, rowsep=1,
                          colspecs=None):
        layout = ET.SubElement( parent, "tgroup", cols=str(nbcols),
                                align=align, colsep=str(colsep),
                                rowsep=str(rowsep) )
        if colspecs:
            for i, colspec in enumerate(colspecs):
                ET.SubElement( layout, "colspec", colname="c%s"%i, colwidth=colspec )
        return layout

    def table_cell_node(self, parent, align='', value=u''):
        """ create a DOM node <entry> """
        entry = ET.SubElement(parent, 'entry')
        if align and value:
            entry.set('align', align)
            entry.text = value

    def section(self, parent, title, id=None):
        section = ET.SubElement(parent, "section")
        if id:
            section.set("id",id)
        assert isinstance(title,unicode)
        ET.SubElement(section,"title").text = title
        return section

    def para(self, parent, text):
        assert isinstance(text, unicode)
        ET.SubElement(parent, "para").text = text

    def formalpara(self, parent, title, id=None):
        para = ET.SubElement(parent,"formalpara")
        if id:
            para.set("id",id)
        assert isinstance(title, unicode)
        ET.SubElement(para,"title").text = title
        return para

# other utilities and abstract classes ########################################

# FIXME handle english (lang='en')

TVA = 19.6
EXT_DATE_FORMAT = u'%Y-%m-%d'
FULL_DATE_FORMAT = u'%d/%m/%Y'
DATE_NOT_SPECIFIED = "non spécifié"
TOTAL_DATE = u"L'ensemble du projet se déroule entre le %s et le %s."
TOTAL_DURATION = u"La charge totale se chiffre à %s."
TOTAL_DURATION_UNIT = u"%.1f jour.homme"
TOTAL_DURATION_UNITS = u"%.1f jours.homme"

def get_daily_labor(number):
    """return a string with unit jour(s).homme"""
    if number <= 1.:
        return TOTAL_DURATION_UNIT % number
    else:
        return TOTAL_DURATION_UNITS % number

class CostData:
    """handle global calculation: cost, duration, ressources' rate
    """

    def __init__(self, projman):
        self.projman = projman
        self.project_cost = 0.0
        self.project_duration = 0.0
        self._used_resources = set()
        self.compute(projman.root_task)

    def compute(self, project):
        for task in project.children:
            self._compute(task)

    def _compute(self, task, level=0):
        try:
            task_cost = self.projman.get_task_total_cost(task.id, task.duration)
        except KeyError:
            task_cost = 0
        self.project_cost += task_cost
        # compute global duration
        self.project_duration += task.duration
        # set used_resources for legend
        grouped = self.projman.costs.groupby('task', 'resource')
        # grouped[task.id] is a dictionnary (res_id/rows)
        self._used_resources |= set(grouped.get(task.id, []))
        for each in task.children:
            self._compute(each, level+1)

    def used_resources(self):
        return [self.projman.get_resource(rid) for rid in self._used_resources if rid]


class XMLView:
    name = None

    def __init__(self, config):
        self.config = config
        self.max_level =  self.config.level

    def unique_id(self, nid):
        # use getattr since not all commands support task-root option
        vtask_root = self.config.task_root
        if vtask_root:
            return '%s-%s' % (vtask_root, nid)
        return nid

    def generate(self, xmldoc, projman):
        """return a dr:object node for the rate section view"""
        self._init(projman, xmldoc)
        root = xmldoc.getroot()
        obj = self.dbh.object_node(root, self.unique_id(self.name))
        self.add_content_nodes(obj)
        return obj
    
    def _init(self, projman, xmldoc=None, dbh=None):
        """initialize view members necessary for content generation"""
        self.projman = projman
        if self.config.task_root == None:
            self.config.task_root = self.projman.root_task.id
        self.dbh = dbh or DocbookHelper()
        try:
            self.cdata = projman.__view_cost_data
        except AttributeError:
            self.cdata = projman.__view_cost_data = CostData(projman)

    def subview_content_nodes(self, parent, viewklass):
        """instantiate the given view class and return its content nodes"""
        view = viewklass(self.config)
        view._init(self.projman, dbh=self.dbh)
        view.add_content_nodes( parent )

    def add_content_nodes(self, parent):
        raise NotImplementedError

# actual views ################################################################

class RatesSectionView(XMLView):
    name = 'rates-section'
    
    def add_content_nodes(self, parent):
        section = self.dbh.section(parent, u"Tarifs journaliers", id=self.unique_id('rate-section'))
        self.dbh.para(section, u"Coût pour une journée type de travail:")
        resources = self.cdata.used_resources()
        self.add_resources_rates(section, resources)

    def add_resources_rates(self, parent, resources):
        """ create a DOM node <itemizedlist> containing the legend of table"""
        list_items = ET.SubElement(parent, "itemizedlist")
        task_types = []
        hourly_cost = [] # used only in case of old resources type definition
        for task in self.projman.root_task.leaves():
            if task.task_type: #new resources role definition
                role = self.projman.resource_role_set.get_resource_role(task.task_type)
                if not role in task_types:
                    task_types.append(role)
        if not task_types: # old resource type definition
            for res in self.projman.get_resources():
                resource = self.projman.get_resource(res)
                if not resource.type in task_types:
                    task_types.append(resource.type)
                    hourly_cost.append(resource.hourly_rate[0])
        for i, role in enumerate(task_types):
            if type(role) ==str:
                r_info = '%s : %s' %(role, format_monetary(hourly_cost[i] * HOURS_PER_DAY))
            else:
                r_info = '%s (%s) : %s' %(role.name, role.id, format_monetary(role.hourly_cost * HOURS_PER_DAY))     
            item = ET.SubElement(list_items, "listitem")
            self.dbh.para( item, r_info )


class DurationSectionView(XMLView):
    name = 'duration-section'

    def add_content_nodes(self, parent):
        section = self.dbh.section(parent, u"Durée totale", id=self.unique_id(u"duration-section"))
        self.subview_content_nodes(section, DateParaView)
        self.subview_content_nodes(section, DurationParaView)

class DateParaView(XMLView):
    name = 'dates-para'

    def add_content_nodes(self, parent):
        begin, end = self.projman.get_task_date_range(self.projman.root_task)
        text = TOTAL_DATE % (begin.strftime(FULL_DATE_FORMAT),
                             end.strftime(FULL_DATE_FORMAT))
        self.dbh.para(parent, text)
        ET.SubElement(parent,"para").text = text

class DurationParaView(XMLView):
    name = 'duration-para'

    def add_content_nodes(self, parent):
        text = TOTAL_DURATION % get_daily_labor(self.projman.root_task.maximum_duration())
        ET.SubElement(parent,"para").text = text

class CostTableView(XMLView):
    name = 'cost-table'
    ENTETE = u"Tableau récapitulatif des coûts."

    def add_content_nodes(self, parent):
        """return a dr:object node for the cost table view"""
        self.projman.update_caches()
        table = ET.SubElement(parent,"table")
        ET.SubElement(table, "title").text = self.ENTETE
        # fill column information for table
          # find resources roles
        root = self.projman.get_task(self.config.task_root)
        resources = self.projman.get_resources()
        set_res = []
        for res in resources:
            set_res.append( self.projman.get_resource(res)) 
        root_resources = root.get_linked_resources(set_res)
        self.set_res = []
        for i in range(len(root_resources)):
            res = root_resources.pop()
            self.set_res.append(res)
        self.roles = []
        for task in self.projman.root_task.leaves():
            if task.task_type: #according to new definition of roles resources
                role_ = self.projman.resource_role_set.get_resource_role(task.task_type)
                if not role_ in self.roles:
                    self.roles.append(role_)
        if self.roles == []: # old definition of resources type
            for res in self.set_res:
                resource = self.projman.get_resource(res)
                if not resource.type in self.roles:
                    self.roles.append(resource.type)
        len_ = len(self.roles)
        specs = [u'3*']
        for i in range(len_):
            specs.append(u'1*')
        specs.append(u'1*')
        
        # create table 
        layout = self.dbh.table_layout_node(table, len_+2, colspecs=specs)
        head = self.table_head(layout, len_+2)
        # table body
        tbody = ET.SubElement(layout, "tbody")
        self.set_tbody(tbody)
        for child in self.projman.root_task.children:
            if child.TYPE == 'task' and child.level:
                self.color = 0
                self._build_task_node(tbody, child, child.level)

    def set_tbody(self, tbody):
        tbody.set(LDG_NS+"row-borders", 'false')
        tbody.set(LDG_NS+'row-backgrounds', "never")

    def set_row(self, row):
        row.set(LDG_NS+'border-top', "true")
        row.set(LDG_NS+'bold', "true")

    def table_head(self, parent, len_):
        """ create a DOM node <thead> """
        thead = ET.SubElement(parent, 'thead')
        row =  ET.SubElement(thead,'row')
        self.dbh.table_cell_node(row, 'left', u'Tâches')
        for i in range(len_ - 2):
            role = self.roles[i]
            if type(role) == str:
                self.dbh.table_cell_node(row, 'center', u'Charge "%s"' %role)
            else:
                self.dbh.table_cell_node(row, 'center', u'Charge "%s"' %role.id)
        self.dbh.table_cell_node(row, 'right', u'Coût (euros)')
        return thead

    def _build_task_node(self, tbody, task, level=1):
        """format a task in as a row in the table"""
        if not task.children:
            self.row_element(tbody, task, level)
        elif task.children and level <= self.max_level:
            self.empty_row_element(tbody, task, level)
            for child in task.children:
                if child.TYPE == 'milestone':
                    continue
                if task.level == 1:
                    self.color += 1
                self._build_task_node(tbody, child, level+1)
            self.synthesis_row_element(tbody, task, level)
        else:
            self.synthesis_row_element(tbody, task, level)

    def row_element(self, tbody, task, level=1):
        """ create a DOM element <row> with values in task node"""
        row = ET.SubElement(tbody, 'row')
        if self.color % 2:
            row.set(LDG_NS+'background', "true")
        indent = u'\xA0 '*(level-1)*2
        costs, durations = self.projman.get_task_costs(task.id, task.duration)
        # task title
        if level == 1:
            self.set_row(row)
            self.dbh.table_cell_node(row, 'left', indent+task.title)
        else:
            self.dbh.table_cell_node(row, 'left', indent+PUCE+task.title)
        # task duration
        for role in self.roles:
            if task.children:
                self.dbh.table_cell_node(row)
            else:
                duration = 0
                for res in durations:
                    resource = self.projman.get_resource(res)
                    if not type(role) == str: # according to new resqoiurces definition
                        if role.id in resource.id_role:
                            duration = durations[res]                                
                            self.dbh.table_cell_node(row, 'center', "%s" %duration)
                    else: #old definition of resource type
                        if role == resource.type:
                            duration = durations[res]                                
                            self.dbh.table_cell_node(row, 'center', "%s" %duration)
                if duration == 0:
                    self.dbh.table_cell_node(row)
        # task cost by resources
        if task.children:
            self.dbh.table_cell_node(row)
        else:
            cost = 0
            for res in costs:
                cost += costs[res]
            self.dbh.table_cell_node(row,'right', u'%s' %format_monetary(cost))
        return row

    def empty_row_element(self, tbody, task, level=0):
        """ create a DOM element <row> with values in task node"""
        row =  ET.SubElement(tbody, 'row')
        if self.color % 2:
            row.set(LDG_NS+'background', "true")
        if level == 1:
            self.set_row(row)
        indent = u'\xA0 '*(level-1)*2
        # task title
        if level == 1:
            self.dbh.table_cell_node(row, 'left', indent+task.title)
        else:
            self.dbh.table_cell_node(row, 'left', indent+PUCE+task.title)
        for role in self.roles:
            self.dbh.table_cell_node(row)
        self.dbh.table_cell_node(row)
        return row

    def synthesis_row_element(self, tbody, task, level):
        durations_ = {}
        # task title
        row = ET.SubElement(tbody, 'row')
        if task.level <= self.max_level and task.children:
            row.set(LDG_NS+'italic', "true")
            string = u'Synthèse '
            if task.level == 1:
                 row.set(LDG_NS+'border-bottom', "true")
                 row.set(LDG_NS+'bold', "true")
        else:
            string = u''
        if  self.color % 2 and task.level > 1: 
            row.set(LDG_NS+'background', "true")
        indent = u'\xA0 '*(level-1)*2
        self.dbh.table_cell_node(row, 'left', indent+string+task.title)
        for res in self.set_res:
            durations_.setdefault(res,0)
        durations_ = {}
        costs_ = 0
        for child in task.children: # ?? leaves() should be better
            if child.children and child.level > self.max_level:
                raise ViewException('task %s must have no child to generate views' %child.id)
            costs, durations = self.projman.get_task_costs(child.id, child.duration)
            for children in child.children:
                costs_child, durations_child = self.projman.get_task_costs(children.id, children.duration)
                for res in durations_child:
                    if not res in durations_:
                        durations_.setdefault(res, 0)
                    durations_[res] += durations_child[res]
                    costs_ += costs_child[res]
            for res in durations:
                if not res in durations_:
                    durations_.setdefault(res, 0)
                durations_[res] += durations[res]
                costs_ += costs[res]
        for role in self.roles:
            duration = 0
            for res in durations_:
                resource = self.projman.get_resource(res)
                if not type(role) == str: # according to new resqoiurces definition
                    if role.id in resource.id_role:
                        duration = durations_[res]                                
                        self.dbh.table_cell_node(row, 'center', "%s" %duration)
                else: #old definition of resource type
                    if role == resource.type:
                        duration = durations_[res]                                
                        self.dbh.table_cell_node(row, 'center', "%s" %duration)
            if duration == 0:
                self.dbh.table_cell_node(row)
        self.dbh.table_cell_node(row,'right', u'%s' %format_monetary(costs_))
        
class CostParaView(XMLView):
    name = 'cost-para'
    TOTAL_COST = u"Le coût total se chiffre à %s euros HT, soit %s euros TTC en appliquant les taux actuellement en vigueur."

    def add_content_nodes(self, parent):
        """return a dr:object node for the cost paragraph view"""
        cost = self.cdata.project_cost
        text = self.TOTAL_COST % (format_monetary(cost),
                                  format_monetary(cost * (1+TVA/100)))
        ET.SubElement(parent, 'para').text = text

class TasksListSectionView(XMLView):
    name = 'tasks-list-section'

    def add_content_nodes(self, parent):
        """return a dr:object node for the tasks list section view"""
        self.projman.update_caches()
        for child in self.projman.root_task.children:
            if not child.TYPE == 'milestone':
                self._build_task_node(parent, child)

    def _build_task_node(self, parent, task):
        section = self.dbh.section(parent, task.title, id=task.id)
        # fill description
        if task.description != "":
            # create xml-like string
            # encode it and create XML tree from it
            # FIXME !!!
            assert isinstance(task.description, unicode), task.description
            desc = "<?xml version='1.0' encoding='UTF-8'?><para>%s</para>" \
                   % task.description.encode('utf8')
            try:
                description_doc = ET.fromstring(desc)
            except Exception, exc:
                print desc
                raise
            section.append( description_doc )
        if task.children:
            self._build_tables(task,section)
            self.add_para_total_load(section, task)
            if  self.config.display_dates:
                self.add_dates(section, task)
            # print children
            for child in task.children:
                if not  child.TYPE == 'milestone':
                    self._build_task_node(section, child)
        else:
            # task is a real task (leaf without any sub task)
            
            # add resources' info
            if task.TYPE == 'task' :
                self.resource_node(section, task)

            # add date constraints
            if self.config.display_dates:
                self.add_dates(section, task)
            if task.link:
                para = self.dbh.formalpara(section, u'Voir aussi')
                link = ET.SubElement(para, "ulink")
                link.set(u'url', '%s' %task.link)
        return section

    def add_dates(self, parent, task):
        """print begin and end of  task"""
        para = self.dbh.formalpara(parent,u'Dates')
        para = ET.SubElement(parent,"para")
        list_ = ET.SubElement(para, "itemizedlist")
        debut, fin = self.projman.get_task_date_range(task)
        item = ET.SubElement(list_,'listitem')
        self.dbh.para(item, u"Date de début : %s" %debut.Format(EXT_DATE_FORMAT))
        item = ET.SubElement(list_,'listitem')
        self.dbh.para(item, u"Date de fin :   %s" %fin.Format(EXT_DATE_FORMAT))
            
    def add_para_total_load(self, parent, task):
        """print total load (load for each resources)"""
        para = self.dbh.formalpara(parent,u'Charge totale')
        para = ET.SubElement(parent,"para")
        list_ = ET.SubElement(para, "itemizedlist")
        durations = {}
        costs = {}
        for leaf in task.leaves():
            if leaf.TYPE == 'milestone':
                continue
            costs_, durations_ = self.projman.get_task_costs(leaf.id, leaf.duration)
            for res in durations_:
                if not res in durations:
                    durations.setdefault(res, 0)
                    costs.setdefault(res, 0)
                durations[res] += durations_[res]
                costs[res] += costs_[res]
        for res in durations:
            resource = self.projman.get_resource(res)
            
            #if self.projman.resource_role_set.width() >1:  #use new definition of resources
                #for leaf in task.leaves(): # does not work
            #parcours des feuilles pour connaitre la liste des roles, le tps consomme
            #est donne par duration
            # on ne peut pas retrouver le role-type d'une tache a partir d'unr resource,
            # car elle peut avoir plusieurs roles. On suppose donc que les feuilles
            # auront le meme role-type
            role = None
            for leaf in task.leaves():
                if leaf.task_type or task.task_type:
                    if leaf.TYPE == 'milestone':
                        continue
                    role_ = self.projman.resource_role_set.get_resource_role(leaf.task_type)
                    role = role_.name
#            else:  # use old definition
            if not(role):
                resource = self.projman.get_resource(res)
                role = resource.type
            item = ET.SubElement(list_,'listitem')
            self.dbh.para(item, u"%s (%s) : %s jour.hommes" %(role, resource.name, durations[res]))
             
            
    def resource_node(self, parent, task):
        """ create a DOM node
        <formalpara>
          <title>Charge et répartition</title>

          <para>
            <itemizedlist>
              <listitem><para>role (res_id) : duration jours.homme</para></listitem>
              <listitem><para>   ...    </para></listitem>
            </itemizedlist>
          </para>
        </formalpara>
        """
        # use new resources definition:
        para = self.dbh.formalpara(parent,u'Charge totale')
        para = ET.SubElement(parent,"para")
        list_ = ET.SubElement(para, "itemizedlist")
        _, duration = self.projman.get_task_costs(task.id, task.duration)
        duration_ = {}
        for res in duration:
                resource = self.projman.get_resource(res)
                if task.task_type:  # use new resources definition
                    res_role =  self.projman.resource_role_set.get_resource_role(task.task_type)
                    role = res_role.name
                else:  # use old definition of resource roles
                    role = resource.type
                item = ET.SubElement(list_,'listitem')
                self.dbh.para(item, u"%s (%s) : %s " %(role, resource.name, duration[res]))
        return para

    def _build_tables(self, task, section):
        """ build tables of principal information of the task
        one for duration of subtasks and date of end of subtasks
        and another to describe deliverables"""
        # table of subtasks
        table = ET.SubElement(section,"informaltable")
        layout = self.dbh.table_layout_node(table, 2, colspecs=('2*', '1*'))
        self.table_head_task(layout)
          # table body
        tbody = ET.SubElement(layout, "tbody")
          # ligns of table
        for child in task.children:
            if not child.TYPE == 'milestone':
                self.row_element(child, tbody)
        # table of liverables :
        # find milestones
        milestones = []
        for child in task.children:
            if child.TYPE == 'milestone':
                milestones.append(child)
        if milestones:
            table = ET.SubElement(section,"informaltable")
            if self.config.display_dates:
                layout = self.dbh.table_layout_node(table, 2, colspecs=('2*', '1*'))
            else:
                layout = self.dbh.table_layout_node(table, 1, colspecs=('1*',))
            self.table_head_milestone(layout)
            # table body
            tbody = ET.SubElement(layout, "tbody")
        for mil_ in milestones:
            row = ET.SubElement(tbody, 'row')
            self.dbh.table_cell_node(row, 'left', u'%s' %mil_.title)
            if self.config.display_dates:
                # find end of tasks
                _, end = self.projman.get_task_date_range(mil_)
                self.dbh.table_cell_node(row, 'left', u'%s' %end.Format(EXT_DATE_FORMAT))
        
    def table_head_task(self, parent):
        """ create a DOM node <thead> for the task table """
        thead = ET.SubElement(parent, 'thead')
        row = ET.SubElement(thead,'row')
        self.dbh.table_cell_node(row, 'left', u'Tâches contenues')
        self.dbh.table_cell_node(row, 'left', u'Charge')
        return thead

    def table_head_milestone(self, parent):
        """ create a DOM node <thead> for the milestone table """
        thead = ET.SubElement(parent, 'thead')
        row = ET.SubElement(thead,'row')
        self.dbh.table_cell_node(row, 'left', u'Jalons')
        if  self.config.display_dates:
            self.dbh.table_cell_node(row, 'left', u'Date de livraison')
        return thead

    def row_element(self, task, tbody):
        """ create a DOM element <row> with values in task node"""
        if task.children:
            row = ET.SubElement(tbody, 'row')
            # task title
            self.dbh.table_cell_node(row, 'left', task.title)
            # task duration and role of resources
            duration = 0
            resources = set()
            for child in task.children:
                if child.TYPE == 'milestone':
                    continue
                duration += child.duration
                if child.task_type: #use new project description
                    res_type = child.task_type
                    # get role title
                    res_role = self.projman.resource_role_set.get_resource_role(res_type)
                    role = res_role.name
                    resources.add(role)
                else: # use old project description
                    res_ = child.get_resources()
                    res = res_.pop()
                    res = self.projman.get_resource(res)
                    role = res.type
                    resources.add(role)
            string = u'%s' %resources.pop()
            for role in resources:
                string += ', %s' %role
            self.dbh.table_cell_node(row, 'left',string +u' : %s j.h' %duration)
        else:
            row = ET.SubElement(tbody, 'row')
            # task title
            self.dbh.table_cell_node(row, 'left', task.title)
            # task duration and role of resources
            duration = task.duration and unicode(task.duration) or u''
            if task.task_type: #use new project description
                res_type = task.task_type
                # get role title
                res_role = self.projman.resource_role_set.get_resource_role(res_type)
                role = res_role.name
            else: # use old project description
                resources = task.get_resources()
                res_id = resources.pop() # en pratique, on n'associe jamais
                                         # des resources avec des competences differentes
                                         # sur une meme tache ( d'ou pop)
                res = self.projman.get_resource(res_id)
                role = res.type
            self.dbh.table_cell_node(row, 'left', u'%s : %s j.h' %(role, duration))
            # compute end of the task (used in second table)
            return row
    
class DurationTableView(CostTableView):
    name = 'duration-table'
    ENTETE = u"Tableau récapitulatif des dates."

    def add_content_nodes(self, parent):
        """return a dr:object node for the cost table view"""
        self.projman.update_caches()
        table = ET.SubElement(parent,"table")
        ET.SubElement(table,"title").text = self.ENTETE
        # fill column information for table
        layout = self.dbh.table_layout_node(table, 3, colspecs=('3*', '1*', '1*'))
        self.table_head(layout)
        # table body
        tbody = ET.SubElement(layout, "tbody")
        # add tbody attributes
        self.set_tbody(tbody)        
        for child in self.projman.root_task.children:
            if child.TYPE == 'task':
                self.color = 0
                self._build_task_node(tbody, child, child.level)

    def synthesis_row_element(self, row, task, level):
        begin, end = self.projman.get_task_date_range(task)
        #row =  ET.SubElement(parent, 'row')
        # indentation
        indent = u'\xA0 '*(level-2)*2
        # task title
        self.dbh.table_cell_node(row, 'left', indent+u'Synthèse ' + task.title)
        # task begin & end
        date_begin, date_end = self.projman.get_task_date_range(task)
        self.dbh.table_cell_node(row, 'center', date_begin.date)
        self.dbh.table_cell_node(row, 'center', date_end.date)
        return row

    def _build_task_node(self, tbody, task, level=0):
        """format a task in as a row in the table"""
        row = ET.SubElement(tbody, 'row')
        if self.color % 2: 
            row.set(LDG_NS+'background', "true")
        if level == 1:
            self.set_row(row)
        if task.children and level < self.max_level:
            self.row_empty_element(row, task, level)
            for child in task.children:
                if child.TYPE == 'milestone':
                    continue
                if task.level == 1:
                    self.color += 1
                self._build_task_node(tbody, child, level+1)
            row = ET.SubElement(tbody, 'row')
            if self.color % 2 and level > 1:
                row.set(LDG_NS+'background', "true")
            row.set(LDG_NS+'italic', "true")
            if task.level == 1:
                row.set(LDG_NS+'border-bottom', "true")
                row.set(LDG_NS+'bold', "true")
            self.synthesis_row_element(row, task, level+1)
        else:
            #row.set(LDG_NS+'border-bottom', "true")            
            self.row_element(row, task, level)
            
    def table_head(self, table):
        """ create a DOM node <thead> """ 
        thead = ET.SubElement(table, "thead")
        row = ET.SubElement(thead,"row")
        self.dbh.table_cell_node(row, 'leaf', u'Tâches')
        self.dbh.table_cell_node(row, 'center', u'Date de début')
        self.dbh.table_cell_node(row, 'center', u'Date de fin')
        return thead

    def row_element(self, row, task, level=0):
        """ create a DOM element <row> with values in task node"""
        #row = ET.SubElement(tbody, 'row')
        # indentation
        indent = u'\xA0 '*(level-1)*2
        # task title
        if level == 1:
            self.dbh.table_cell_node(row, 'left', indent+task.title)
        else:
            self.dbh.table_cell_node(row, 'left', indent+PUCE+task.title)
        # task begin & end
        date_begin, date_end = self.projman.get_task_date_range(task)
        self.dbh.table_cell_node(row, 'center', date_begin.date)
        
        self.dbh.table_cell_node(row, 'center', date_end.date)
        return row

    def row_empty_element(self, row, task, level=0):
        """ create a DOM element <row> with values in task node"""
        #row = ET.SubElement(tbody, 'row')
        # indentation
        indent = u'\xA0 '*(level-1)*2
        # task title
        if level == 1:
            self.dbh.table_cell_node(row, 'left', indent+task.title)
        else:
            self.dbh.table_cell_node(row, 'left', indent+PUCE+task.title)
        # task begin & end
        self.dbh.table_cell_node(row)
        self.dbh.table_cell_node(row)
        return row

ALL_VIEWS = {}
for klass in (RatesSectionView,
              DurationTableView, DurationParaView, DurationSectionView,
              DateParaView,
              CostTableView, CostParaView,
              TasksListSectionView):
    ALL_VIEWS[klass.name] = klass
    



# -*- coding: iso-8859-1 -*-
# Copyright (c) 2000-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
schedule project without caring about resource conflicts
"""

__revision__ = "$Id: dumb.py,v 1.3 2005-11-10 11:34:21 arthur Exp $"

from mx.DateTime import today

from projman.lib.constants import *

class DumbScheduler:
    """
    schedule project without caring about resource conflicts
    nor task constraints
    """
    
    def __init__(self, project):
        self.project = project
        
    def _process_node(self, node):
        # get dates
        begin, end = None, None
        for c_type, date in node.get_date_constraints():
            if c_type == BEGIN_AFTER_DATE :
                begin = date # FIXME what if more than one ?
            elif c_type == END_BEFORE_DATE :
                end = date # FIXME what if more than one ?
        if not begin and not end:
            begin = today()

        # add activities
        activities = []
        resources = [self.project.get_resource(r_id)
                     for r_type, r_id, usage in node.get_resource_constraints()]
        if node.TYPE=="milestone":
            self.project.milestones[node.id] = begin
            return []
        if not resources:
            print "WARNING: task %s has no resource and will not be scheduled"%node.id
        days = {}
        duration = node.duration*(1-node.progress)
        if begin:
            for resource in resources:
                days[resource] = begin
            while duration and resources:
                for resource in resources:
                    date = days[resource]
                    while not resource.is_available(date):
                        date += 1
                    # usage = 1
                    activities.append( (date, date, resource.id, node.id, 1) )
                    days[resource] = date + 1
                duration -= 1
        else:
            for resource in resources:
                days[resource] = end
            while duration and resources:
                for resource in resources:
                    date = days[resource]
                    while not resource.is_available(date):
                        date -= 1
                    # usage = 1
                    activities.append( (date, date, resource.id, node.id, 1) )
                    days[resource] = date - 1
                duration -= 1
        return activities
    
                
    def schedule(self, verbose=0, **kw):
        """
        Update the project's schedule
        Return list of errors occured during schedule
        """
        activities = []
        for leaf in self.project.root_task.leaves():
            activities.extend(self._process_node(leaf))
        self.project.add_schedule(activities)
        return []




# -*- coding: iso-8859-1 -*-

import os, sys

from logilab.common import testlib 

class SmokeTC(testlib.TestCase):
    CMD = 'projman'

    ## RESULT_DIR contiendra tous les fichiers et images générés lors des tests
    RESULT_DIR = 'smoke_result'

    ## FILE_ASSEMBLAGE est un input utlisant l'assemblage de projets
    FILE_ASSEMBLAGE = 'projman/example.xml'

    def setUp(self):
        # check that path specified might be ok
        dir = os.path.abspath(sys.argv[1])
        os.chdir(dir)
        list_dirs = os.listdir('.')
        if 'projman' not in list_dirs:
            print "ERROR:"
            print "You must specify as parameter the root directory of all examples."
            print "Tipically, ~/cvs/public/soft/projman/examples"
            sys.exit()

        ## create dir for output
        if RESULT_DIR not in list_dirs:
            print "creating directory", RESULT_DIR, "for resulting files"
            os.mkdir(RESULT_DIR)

        ## change dir to output
        os.chdir(dir + '/' + RESULT_DIR)

    def ask_ok(self) :
        r = raw_input('ok [Y]/No ?')
        if r.lower() in ('n', 'no') :
            sys.exit()

    def test_gantt(self):
        "diagramme gantt..."
        os.system(CMD+' --diagram --diagram-type gantt --timestep 7 proj.xml gantt.png')
        os.system('display gantt.png')
        self.ask_ok()

    def test_ressources(self):
        "diagramme ressources..."
        os.system(CMD+' --diagram --diagram-type resources --timestep 7 proj.xml resource.png')
        os.system('display resource.png')
        self.ask_ok()

    def test_gantt_ressources(self):
        "diagramme gantt+ressources..."
        os.system(CMD+' --diagram --diagram-type gantt-resources --timestep 7 proj.xml gantt-resource.png')
        os.system('display gantt-resource.png')
        self.ask_ok()

    def test_ordonnancement(self):
        "ordonnancement..."
        os.system(CMD+' --plan --include-references proj.xml plan.xml')
        self.ask_ok()

    def test_gantt2(self):
        "diagramme gantt..."
        os.system(CMD+' --diagram --diagram-type gantt --timestep 7 proj.xml gantt.png')
        os.system('display gantt.png')
        self.ask_ok()

    def test_gantt_ressources2(self):
        "diagramme gantt+ressources..."
        os.system(CMD+' --diagram --diagram-type gantt-resources --timestep 7 proj.xml gantt-resource.png')
        os.system('display gantt-resource.png')
        self.ask_ok()

    def test_xmldoc(self):
        "xmldoc tasks-lists..."
        os.system(CMD+' --xml-doc --view=tasks-list proj.xml tasks_list.xml')
        self.ask_ok()
        os.system('mkdoc --target=pdf --stylesheet=standard tasks_list.xml')
        os.system('xpdf tasks_list.pdf')

    def test_costs(self):
        "xmldoc tasks-costs..."
        os.system(CMD+' --xml-doc --view=tasks-costs proj.xml tasks_costs.xml')
        self.ask_ok()
        os.system('mkdoc --target=pdf --stylesheet=standard tasks_costs.xml')
        os.system('xpdf tasks_costs.pdf')

    def test_dates(self):
        "xmldoc tasks-dates..."
        os.system(CMD+' --xml-doc --view=tasks-dates proj.xml tasks_dates.xml')
        self.ask_ok()
        os.system('mkdoc --target=pdf --stylesheet=standard tasks_dates.xml')
        os.system('xpdf tasks_dates.pdf')


    def test_gantt4(self):
        "diagramme gantt..."
        os.system(CMD+' --diagram --diagram-type gantt --timestep 7 proj.xml gantt_a.png')
        os.system('display gantt_a.png')
        self.ask_ok()

    def test_ressources3(self):
        "diagramme ressources..."
        os.system(CMD+' --diagram --diagram-type resources --timestep 7 proj.xml resource_a.png')
        os.system('display resource_a.png')
        self.ask_ok()

    def test_gantt_ressources3(self):
        "diagramme gantt+ressources..."
        os.system(CMD+' --diagram --diagram-type gantt-resources --timestep 7 proj.xml gantt-resource_a.png')
        os.system('display gantt-resource_a.png')
        self.ask_ok()

    def test_ordonnancement2(self):
        "ordonnancement..."
        os.system(CMD+' --plan --include-references proj.xml plan.xml')
        self.ask_ok()

    def test_gantt5(self):
        "diagramme gantt..."
        os.system(CMD+' --diagram --diagram-type gantt --timestep 7 proj.xml gantt.png')
        os.system('display gantt.png')
        self.ask_ok()

    def test_gantt6(self):
        "diagramme gantt+ressources..."
        os.system(CMD+' --diagram --diagram-type gantt-resources --timestep 7 proj.xml gantt-resource.png')
        os.system('display gantt-resource.png')
        self.ask_ok()



if __name__ == '__main__':
    testlib.unittest_main()



# -*- coding: iso-8859-1 -*-
"""
unit tests for module projman.lib.Task

Projman - (c)2000-2002 LOGILAB <contact@logilab.fr> - All rights reserved.
Home: http://www.logilab.org/projman

This code is released under the GNU Public Licence v2. See www.gnu.org.
"""

from logilab.common import testlib
from mx.DateTime import DateTime

from logilab.common.compat import set
from logilab.common.testlib import TestCase

from projman.lib import MileStone, Task
from projman.lib.constants import *

class TaskTC(TestCase):
    
    def setUp(self):
        """called before each test from this class

        parent (inge1/1, inge2/1)
          |
          |--- child1 (inge1/1), 10
          |
          |--- child2
          |      |
          |      |--- child2_1 (inge1/0.6), 5 
          |      |
          |      `--- child2_2 (inge2/1), 12
          |
          `--- stone [milestone]
        """
        self.parent = Task('parent')
        self.child1 = Task('child1')
        self.child2 = Task('child2')
        self.child2_1 = Task('child2_1')
        self.child2_2 = Task('child2_2')
        self.stone = MileStone('stone')
        # building tree
        self.parent.append(self.child1)
        self.parent.append(self.child2)
        self.child2.append(self.child2_1)
        self.child2.append(self.child2_2)
        self.parent.append(self.stone)
        # set duration
        self.child1.duration = 10
        self.child2_1.duration = 5
        self.child2_2.duration = 12
        # set of resources constraints
        self.parent.add_resource_constraint('worker', 'inge1')
        self.parent.add_resource_constraint('worker', 'inge2')
        self.child1.add_resource_constraint('worker', 'inge1')
        self.child2_1.add_resource_constraint('worker', 'inge1')
        self.child2_2.add_resource_constraint('worker', 'inge2')
        # set of dates
        self.date_last_week = DateTime(2004, 10, 1)
        self.date_today = DateTime(2004, 10, 7)
        self.date_tomorrow = DateTime(2004, 10, 8)
        self.date_next_week = DateTime(2004, 10, 13)
        self.date_next_score = DateTime(2004, 10, 26)
        # self.short_activity = Activity(100, self.date_today, self.date_next_week)
        # self.long_activity = Activity(100, self.date_today, self.date_next_score)

    def test_get_task(self):
        """check that finds child
        """
        self.assertEquals(self.parent.get_task('child1'), self.child1)
        self.assertEquals(self.parent.get_task('stone'), self.stone)
        self.assertEquals(self.parent.get_task('child2_1'), self.child2_1)
        self.assertEquals(self.parent.get_task('child2_2'), self.child2_2)
    
    def test_get_resource_constraints(self):
        """test rsource is found, even within parents
        """
        # XXX
        # self.assertEquals(self.parent.get_resource_constraints(), [])
        self.assertEquals(self.child1.get_resource_constraints(), set([('worker', 'inge1')]))
        self.assertEquals(self.child2_1.get_resource_constraints(), set([('worker', 'inge1')]))
        self.assertEquals(self.child2_2.get_resource_constraints(), set([('worker', 'inge2')]))
        
    def test_get_resources(self):
        """check that array of all resource id returned
        """
        self.assertEquals(self.parent.get_resources(), set(['inge1', 'inge2']))
        self.assertEquals(self.child1.get_resources(), set(['inge1']))
        self.assertEquals(self.child2.get_resources(), set(['inge1', 'inge2']))
        self.assertEquals(self.child2_1.get_resources(), set(['inge1']))
        self.assertEquals(self.child2_2.get_resources(), set(['inge2']))

    
    def test_get_priority(self):
        """test value returned
        """
        self.assertEquals(self.parent.get_priority(), None)
        self.assertEquals(self.child1.get_priority(), None)
        self.assertEquals(self.child2.get_priority(), None)
        self.assertEquals(self.child2_1.get_priority(), None)
        self.assertEquals(self.child2_2.get_priority(), None)
        self.parent.priority = 2
        self.assertEquals(self.parent.get_priority(), 2)
        self.assertEquals(self.child1.get_priority(), 2)
        self.assertEquals(self.child2.get_priority(), 2)
        self.assertEquals(self.child2_1.get_priority(), 2)
        self.assertEquals(self.child2_2.get_priority(), 2)
        self.child2.priority = 1
        self.assertEquals(self.parent.get_priority(), 2)
        self.assertEquals(self.child1.get_priority(), 2)
        self.assertEquals(self.child2.get_priority(), 1)
        self.assertEquals(self.child2_1.get_priority(), 1)
        self.assertEquals(self.child2_2.get_priority(), 1)
        
        
    def test_is_finished(self):
        """test achievement state
        """
        #input activity is the accomplished activity related to task
        # default: progress=0, duration=10, short_activity=6
        self.assertEquals(self.parent.is_finished(), False)
        self.assertEquals(self.child1.is_finished(), False)
        self.assertEquals(self.child2.is_finished(), False)
        self.assertEquals(self.child2_1.is_finished(), False)
        self.assertEquals(self.child2_2.is_finished(), False)
        # test on progress
        self.child1.progress = 1
        self.assertEquals(self.parent.is_finished(), False)
        self.assertEquals(self.child1.is_finished(), True)
        self.assertEquals(self.child2.is_finished(), False)
        self.assertEquals(self.child2_1.is_finished(), False)
        self.assertEquals(self.child2_2.is_finished(), False)
        self.child2_2.progress = 1
        self.child2_1.progress = 1
        self.assertEquals(self.child1.is_finished(), True)
        self.assertEquals(self.child2.is_finished(), True)
        self.assertEquals(self.child2_1.is_finished(), True)
        self.assertEquals(self.child2_2.is_finished(), True)
##         # test with duration=10 long_activity=20
##         self.assertEquals(self.parent.is_finished(), True)
##         self.assertEquals(self.single_task.is_finished(), True)
##         self.assertEquals(self.child_task.is_finished(), True)
        
        
    def test_remaining_duration(self):
        """check returned number of days, and its right calculation
        according to progress
        """
        # progress=1, duration=10, short_activity=6
        self.assertRaises(AssertionError, setattr, self.parent, 'progress', 1)
        self.assertRaises(AssertionError, setattr, self.child2, 'progress', 1)
        self.child1.progress = .5
        self.child2_1.progress = .5
        self.child2_2.progress = 1.
        self.assertEquals(self.child2_2.remaining_duration(), 0)
        self.assertEquals(self.child2_1.remaining_duration(), 2.5)
        self.assertEquals(self.child2.remaining_duration(), 2.5)
        self.assertEquals(self.child1.remaining_duration(), 5)
        self.assertEquals(self.parent.remaining_duration(), 7.5)
        self.assertEquals(self.child2.progress, 14.5/17)
        self.assertEquals(self.parent.progress, 19.5/27)

    
    def test_maximum_duration(self):
        """test that all children are taken into account
        """
        self.assertEquals(self.parent.maximum_duration(), 27)
        self.assertEquals(self.child1.maximum_duration(), 10)
        self.assertEquals(self.child2.maximum_duration(), 17)
        self.assertEquals(self.child2_1.maximum_duration(), 5)
        self.assertEquals(self.child2_2.maximum_duration(), 12)


class ConsistencyTC(testlib.TestCase):

    def setUp(self):
        self.date1 = DateTime(2005, 01, 01)
        self.date2 = DateTime(2005, 01, 02)
        self.date3 = DateTime(2005, 01, 03)
        self.date4 = DateTime(2005, 01, 04)
        date2 = self.date2
        date3 = self.date3
        self.parent = Task('parent')
        # add a child per constraint type
        self.child_begin_at = Task('child_begin_at')
        self.child_begin_at.add_date_constraint(BEGIN_AT_DATE, date2)
        self.parent.append(self.child_begin_at)
        
        self.child_begin_after = Task('child_begin_after')
        self.child_begin_after.add_date_constraint(BEGIN_AFTER_DATE, date2)
        self.parent.append(self.child_begin_after)

        self.child_begin_before = Task('child_begin_before')
        self.child_begin_before.add_date_constraint(BEGIN_BEFORE_DATE, date2)
        self.parent.append(self.child_begin_before)

        self.child_end_at = Task('child_end_at')
        self.child_end_at.add_date_constraint(END_AT_DATE, date3)
        self.parent.append(self.child_end_at)
        
        self.child_end_after = Task('child_end_after')
        self.child_end_after.add_date_constraint(END_AFTER_DATE, date3)
        self.parent.append(self.child_end_after)

        self.child_end_before = Task('child_end_before')
        self.child_end_before.add_date_constraint(END_BEFORE_DATE, date3)
        self.parent.append(self.child_end_before)

    def test_parent_conflict(self):
        self.parent.add_date_constraint(BEGIN_AT_DATE, self.date1)
        self.assertEquals(self.parent.check_consistency(), [])
        self.parent.add_date_constraint(BEGIN_AT_DATE, self.date2)
        self.assertEquals(len(self.parent.check_consistency()), 1)
        
    def test_parent_begin_at2(self):
        self.parent.add_date_constraint(BEGIN_AT_DATE, self.date2)
        self.assertEquals(self.parent.check_consistency(), [])

    def test_parent_begin_at3(self):
        self.parent.add_date_constraint(BEGIN_AT_DATE, self.date3)
        self.assertEquals(self.parent.check_consistency(), [])

    def test_parent_begin_at4(self):
        self.parent.add_date_constraint(BEGIN_AT_DATE, self.date4)
        self.assertEquals(self.parent.check_consistency(), [])

    # FIXME etc.
    
class DateConstraintsTC(testlib.TestCase):

    def setUp(self):
        self.parent = Task('parent')
        self.child = Task('child')
        self.parent.append(self.child)
        
    def test_parent_begin_at(self):
        self.assertEquals(self.child.get_date_constraints(), set())
        self.parent.add_date_constraint(BEGIN_AT_DATE, DateTime(2005, 01, 01))
        constraints = self.child.get_date_constraints()
        expected = set([(BEGIN_AFTER_DATE, DateTime(2005, 01, 01), 1)])
        self.assertEquals(constraints, expected)

    def test_parent_begin_after(self):
        self.assertEquals(self.child.get_date_constraints(), set())
        self.parent.add_date_constraint(BEGIN_AFTER_DATE, DateTime(2005, 01, 01))
        constraints = self.child.get_date_constraints()
        expected = set([(BEGIN_AFTER_DATE, DateTime(2005, 01, 01), 1)])
        self.assertEquals(constraints, expected)

    def test_parent_end_at(self):
        self.assertEquals(self.child.get_date_constraints(), set())
        self.parent.add_date_constraint(END_AT_DATE, DateTime(2005, 01, 01))
        constraints = self.child.get_date_constraints()
        expected = set([(END_BEFORE_DATE, DateTime(2005, 01, 01), 1)])
        self.assertEquals(constraints, expected)

    def test_parent_end_before(self):
        self.assertEquals(self.child.get_date_constraints(), set())
        self.parent.add_date_constraint(END_BEFORE_DATE, DateTime(2005, 01, 01))
        constraints = self.child.get_date_constraints()
        expected = set([(END_BEFORE_DATE, DateTime(2005, 01, 01), 1)])
        self.assertEquals(constraints, expected)

class ProgressTC(testlib.TestCase):

    def test_get_progress(self):
        """tests progress' getter"""
        t = Task('hello')
        t.duration = 10
        self.assertEquals(t.progress, 0.)
        #activity = [100, DateTime(2005, 8, 1), DateTime(2005, 8, 4)]
        # 
        #self.assertEquals(t.progress, 0.4)
        t.progress = 0.8
        self.assertEquals(t.progress, 0.8)


    def test_set_progress(self):
        t = Task('hello')
        self.assertRaises(ValueError, setattr, t, 'progress', 12)
        self.assertRaises(ValueError, setattr, t, 'progress', -1)
        t.progress = 1
        self.assertEquals(t.progress, 1.)
    
if __name__ == '__main__':
    testlib.unittest_main()



from logilab.common import testlib

from projman import format_monetary, extract_extension

class Test(testlib.TestCase):

    def test_format_monetary(self):
        self.assertRaises(TypeError, format_monetary, self)
        data = [(121212.01, u"121\xA0212,01"),
                (121000, u"121\xA0000,00"),
                ("121212.01", u"121\xA0212,01"),
                ("1000200.00", u"1\xA0000\xA0200,00"),
                (u"1200.00", u"1\xA0200,00"),
                (u"1200", u"1\xA0200,00"),
                (u"12", u"12,00"),
                (7.9999999999999991, u"8,00"),
                ]
        for input, expected in data:
            self.assertEquals(format_monetary(input), expected)

    def test_format(self):
        self.assertEquals(u"1\xA0200,0", format_monetary("1200.00", nb_decimal=1))
        self.assertEquals(u"1\xA0200,0", format_monetary(1200.00, nb_decimal=1))
        self.assertEquals(u"1\xA0200", format_monetary(u"1200.00", nb_decimal=0))
        self.assertEquals(u"1\xA0200", format_monetary(1200.00, nb_decimal=0))
        self.assertEquals(u"12", format_monetary(u"12", nb_decimal=0))
        self.assertEquals(u"1200,0", format_monetary("1200", nb_decimal=1, space_gap=0))
        self.assertEquals(u"1\xA02\xA00\xA00,00", format_monetary(u"1200", space_gap=1))

    def test_extract_extension(self):
        self.assertEquals(("foo", ""), extract_extension("foo"))
        self.assertEquals(("foo", "txt"), extract_extension("foo.txt"))
        self.assertEquals(("foo", "txt.old"), extract_extension("foo.txt.old"))
        
if __name__ == "__main__":
    testlib.unittest_main()



# -*- coding: utf-8 -*-
# Copyright (c) 2004-2006 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Projman - (c)2000-2006 LOGILAB <contact@logilab.fr> - All rights reserved.

Home: http://www.logilab.org/projman

Manipulate a xml project description.

This code is released under the GNU Public Licence v2. See www.gnu.org.
"""

from logilab.common.tree import VNode 
from mx.DateTime import Time, TimeDelta

from projman import DAY_WEEK

class Calendar(VNode):
    """
    Defines which days are work days.

    All days are work days, unless told otherwise by the calendar.
    """

    TYPE = 'calendar'
    
    def __init__(self, id, name=u''):
        VNode.__init__(self, id)
        self.name = name
        self.day_types = {}
        # example:
        #  {'working': ['DefaultWorking', [(Time(8), Time(12)),
        #                                  (Time(13), Time(17))]
        #              ],
        #  'nonworking': ['DefaultNonworking', []],
        # }
        self.default_day_type = None
        # day type in a usual week
        # ex : {day_of_week:type, ...} and day_of_week
        # in {mon, tue, wed, thu, fri, sat, sun}
        self.weekday = {}
        # list of periods associated to its type 
        #ex: [(from_date, to_date, type), ...]
        # with from_date and to_date as DateTime object
        self.timeperiods = []
        # register the national day, so a relative date 
        # available each year (mm/dd)
        self.national_days = []
        self.start_on = None
        self.stop_on = None

    def add_timeperiod(self, from_datetime, to_datetime, type):
        """
        add a time period in this calendar associated to its type
        """
        new_timeperiods = []
        if self.timeperiods:
            flag = 0
            for timeperiod in self.timeperiods:
                flag2 = 0
                if (timeperiod[1]+1).date >= from_datetime.date \
                       and (timeperiod[1]).date <= to_datetime.date \
                       and timeperiod[2] == type:
                    new_timeperiods.append((timeperiod[0], to_datetime, type))
                    flag = 1
                    flag2 = 1
                elif (timeperiod[1]+1).date >= from_datetime.date \
                     and (timeperiod[1]).date <= to_datetime.date \
                     and (timeperiod[0]).date >= from_datetime.date \
                     and timeperiod[2] == type:
                    new_timeperiods.append((from_datetime, to_datetime, type))
                    flag = 1
                    flag2 = 1
                elif (timeperiod[0]-1).date <= to_datetime.date \
                         and timeperiod[1].date >= to_datetime.date \
                         and timeperiod[2] == type:
                    new_timeperiods.append((from_datetime, timeperiod[1], type))
                    flag = 1
                    flag2 = 1
                elif from_datetime.date <= timeperiod[0].date \
                     and to_datetime.date >= timeperiod[1].date \
                     and timeperiod[2] == type:
                    new_timeperiods.append((from_datetime, to_datetime, type))
                    flag = 1
                    flag2 = 1
                if flag2 == 0:
                    new_timeperiods.append(timeperiod)
            if flag == 0:
                new_timeperiods.append((from_datetime, to_datetime, type))
        else:
            new_timeperiods.append((from_datetime, to_datetime, type))
           
        self.timeperiods = new_timeperiods
        
        # check that timeperiods can merge together
        new_timeperiods = []
        merged_index = []
        for i in range(len(self.timeperiods)):
            flag = 0
            for k in range(len(self.timeperiods)):
                if k != i and k > i:
                    if self.timeperiods[i][0].date == self.timeperiods[k][1].date \
                           and self.timeperiods[i][2] == self.timeperiods[k][2]:
                        new_timeperiods.append(
                            (self.timeperiods[k][0], 
                             self.timeperiods[i][1], 
                             self.timeperiods[i][2]))
                        flag = 1
                        merged_index.append(k)
                    elif self.timeperiods[i][1].date == self.timeperiods[k][0].date \
                             and self.timeperiods[i][2] == self.timeperiods[k][2]:
                        new_timeperiods.append(
                            (self.timeperiods[i][0], 
                             self.timeperiods[k][1], 
                             self.timeperiods[i][2]))
                        flag = 1
                        merged_index.append(k)
            if flag == 0 and i not in merged_index:
                new_timeperiods.append( 
                    (self.timeperiods[i][0], 
                     self.timeperiods[i][1], 
                     self.timeperiods[i][2]))

        self.timeperiods = new_timeperiods
                    

    def availability(self, datetime):
        """
        return work intervals on that datetime.
        """
        if not self.after_start(datetime):
	    return []

	if not self.before_stop(datetime):
	    return []

        if self.is_a_national_day(datetime):
            return []

        daytype = self.get_daytype(datetime)
        return self._get_intervals(daytype)

    def is_a_national_day(self, datetime):
        """
        test if datetime is a national day off
        """
        cal = self
        while isinstance(cal, Calendar):
            if (datetime.month, datetime.day) in cal.national_days:
                return True
            cal = cal.parent
        return False
    
    def after_start(self, datetime):
        cal = self
        while isinstance(cal, Calendar):
            if cal.start_on:
	        if cal.start_on <= datetime:
		    return True
		else:
                    return False
            cal = cal.parent
        return True

    def before_stop(self, datetime):
        cal = self
        while isinstance(cal, Calendar):
            if cal.stop_on:
	        if datetime <= cal.stop_on:
		    return True
		else:
                    return False
            cal = cal.parent
        return True 
 
    def _get_intervals(self, daytype):
        cal = self
        while isinstance(cal, Calendar):
            if daytype in cal.day_types:
                return cal.day_types[daytype][1]
            cal = cal.parent
        raise ValueError('Unknown day type %s' % daytype)

    def get_daytype(self, datetime):
        """
        compute then return day type (a key in self.day_types) for date datetime.
        """
        daytype = self._get_daytype_timeperiods(datetime)
        if daytype is None:
            daytype = self._get_daytype_weekday(datetime)
        if daytype is None:
            daytype = self.get_default_daytype()
        if daytype is None:
            raise Exception('Unable to compute daytype for "%s" in calendar "%s"' 
                            % (datetime, self.name))
        return daytype
    
    def _get_daytype_timeperiods(self, datetime):
        cal = self
        while isinstance(cal, Calendar):
            for from_date, to_date, _type in cal.timeperiods:
                if from_date <= datetime <= to_date:
                    return _type
            cal = cal.parent
        return None

    def _get_daytype_weekday(self, datetime):
        """
        return the type of date acording to the day of the week
        and to the inheriting properties
        """
        day = DAY_WEEK[datetime.day_of_week]
        cal = self
        while isinstance(cal, Calendar):
            if day in cal.weekday:
                return cal.weekday[day]
            cal = cal.parent
        return None

    def get_default_daytype(self):
        cal = self
        while isinstance(cal, Calendar):
            if cal.default_day_type is not None:
                return cal.default_day_type
            cal = cal.parent
        return None

    def get_worktime(self, daytype):
        """
        return a TimeDelta representing the amount of work time on datetime.
        """
        res = TimeDelta(0)
        intervals = self._get_intervals(daytype)
        for from_time, to_time in intervals:
            res += to_time - from_time
        return res






# Copyright (c) 2000-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
base classes for rendering
"""

__revision__ = "$Id: gantt.py,v 1.2 2005-09-08 14:26:06 nico Exp $"

from projman.lib import date_range
from projman.lib.constants import HOURS_PER_DAY
from projman.renderers.abstract import \
     AbstractRenderer, AbstractDrawer, TODAY, \
     TITLE_COLUMN_WIDTH, FIELD_COLUMN_WIDTH, ROW_HEIGHT
from logilab.common.tree import NodeNotFound
from mx.DateTime import oneHour
class GanttRenderer(AbstractRenderer) :
    """
    Render a Gantt diagram
    """
    
    def __init__(self, options, handler, colors_file=None, colors_stream=None) :
        AbstractRenderer.__init__(self, options)
        self.drawer = GanttDrawer(options, handler, colors_file, colors_stream)
        
    def render(self, task, stream):
        """
        render the task as a gantt diagram
        """
        AbstractRenderer.render(self, task, stream)
        self.drawer._handler.save_result(stream)
    
    def _render_body(self, project) :
        """
        generate events to draw a Gantt diagram for the task description
        
        concrete renderer should use the 'write' method or override this method
        to return the generated content.
        """
        self.drawer.main_title('Gantt diagram')
        self.drawer.draw_timeline()
        self.drawer.close_line()

        begin_p, end_p = project.get_task_date_range(project.root_task)
        self.render_node(project.root_task, project, begin_p, end_p)
        for task in self._pending_constraints:
            for c_type, c_id, priority in task.task_constraints:
                try:
                    ct = task.get_node_by_id(c_id)
                except NodeNotFound :
                    ct = None
                if ct and ct in self._visible_tasks:
                    self.drawer.task_constraints(c_type, task, ct, project.factor)        

    def render_node(self, node, project, begin_p, end_p):
        """
        render self and children
        """
        if node.TYPE == 'milestone':
            self.render_milestone(node, project)
        else:
            self.render_task(node, project)
        # hide task under the depth limit
        if self.options.depth and node.depth() >= self.options.depth :
            return
        
        # hide task out of the time line
        if begin_p and end_p:
            if end_p < self.drawer._timeline_days[0] or \
                   begin_p > self.drawer._timeline_days[-1]: 
                return
            
        # render subtasks
        if node.children:
            for node_child in node.children:
                self.render_node(node_child, project, begin_p, end_p)
                
    def render_task(self, task, project):
        """
        generate event for a given task 
        """
        project.get_factor()
        factor = project.factor
        if self.options.del_ended and task.is_finished():
            return
        self.drawer.set_color_set(self._i)
        self._i += 1
        self._visible_tasks[task] = 1
        for val in task.task_constraints:
            if val:
                self._pending_constraints[task] = 1
                break

        self.drawer.open_line()
        self.drawer.main_content(task.title or task.id,
                                 project, task.depth(), task)
        
        if self.options.showids :
            self.drawer.simple_content(task.title)
                
        begin, end = project.get_task_date_range(task)
        end -= oneHour * HOURS_PER_DAY / factor
        self.drawer.task_timeline_bg()
        for day in self.drawer._timeline_days:
            self.drawer.task_timeline(task, True, task.children, '', day,
                                          begin, end, project)
        self.drawer.close_timeline()
        if self.options.rappel:
            self.drawer.main_content(task.title or task.id,
                                     project, task.depth(), task)
        # close table row
        self.drawer.close_line()

    def render_milestone(self, milestone, project):
        """
        generate event for a given task
        """
        self.drawer.set_color_set(self._i)
        self._i += 1
        self._visible_tasks[milestone] = 1
        for val in milestone.task_constraints:
            if val:
                self._pending_constraints[milestone] = 1
                break
        depth = milestone.depth()
        self.drawer.open_line()
        self.drawer.main_content(milestone.title or milestone.id,
                                 project, depth, milestone)

        if self.options.showids :
            self.drawer.simple_content(milestone.title)

        # print task calendar
        for d in self.drawer._timeline_days:
            self.drawer.milestone_timeline(d, milestone, project)
        self.drawer.close_timeline()

        if self.options.rappel:
            self.drawer.main_content(milestone.title or milestone.id,
                                     project, depth, milestone)
        # close table row
        self.drawer.close_line()

class GanttDrawer(AbstractDrawer) :
    """
    Draw a Gantt diagram
    """
    
    ## AbstractDrawer interface #############################################
    
    def _init_table(self):
        """
        initialize fields needed by the table
        """
        AbstractDrawer._init_table(self)
        # mapping to save tasks coordonates
        self._tasks_slots = {}
        # current task
        self._ctask = None
        
    def _get_table_dimension(self, project):
        """
        calculate dimension of the table
        """
        #calculate width
        width = TITLE_COLUMN_WIDTH
        if self.options.rappel:
            width *= 2
        if self.options.showids :
            width += FIELD_COLUMN_WIDTH*1
        if 0 and self.options.detail > 1 :
            width += FIELD_COLUMN_WIDTH*2
        if 0 and self.options.detail > 0 :
            width += FIELD_COLUMN_WIDTH*4
        width += len(self._timeline_days)*self._timestepwidth
        #calculate height
        height = ROW_HEIGHT * (5 + project.get_nb_tasks())
        
        return (width, height)
    
    # project table head/tail #################################################

    def legend(self):
        """
        write the diagram's legend of tasks
        """
        self._legend_task()

    # project table content ###################################################

    def task_timeline(self, task, worked, is_container, text, first_day,
                      begin, end, project):
        """
        write a timeline day for the task (i.e. <timestep> days)
        """
        last_day = first_day + self._timestep - (15 + HOURS_PER_DAY / project.factor) * oneHour
        for d in range(self._timestep):
            for i in range(project.factor):
                day_ = d + first_day + i*(1./project.factor)*HOURS_PER_DAY*oneHour
                if day_.hour >= 12:
                    day_ += oneHour
                worked = False
                if begin and end and begin <= day_ <= end:
                    if is_container or project.is_in_allocated_time(task.id, day_):
                        worked = True
                if task.link and begin == day_:
                    self.open_link(task.link)
                self._task_timeline(worked, is_container,
                                        day_ == begin,
                                        day_ == end,
                                        day_ == first_day,
                                        day_ == end,
                                        day_,
                                        project.factor)
                if task.link and end == day_:
                    self.close_link()
                

    def task_timeline_bg( self ):
        """Draw the background of a timeline"""
        first_day = self._timeline_days[0]
        last_day = self._timeline_days[-1]+self._timestep
        rnge = list( date_range( first_day, last_day ) )
        daywidth = self._daywidth
        self._handler.draw_rect(self._x, self._y+1, daywidth*len(rnge),
                                ROW_HEIGHT-2, fillcolor=self._color_set['WEEKDAY'])
        # draw week-end days-
        for n, day in enumerate(rnge):
            if day.date == TODAY.date :
                bgcolor = self._color_set['TODAY']
            elif day.day_of_week in (5, 6):
                bgcolor = self._color_set['WEEKEND']
            else:
                continue
            self._handler.draw_rect(self._x+n*daywidth, self._y+1, daywidth,
                                    ROW_HEIGHT-2, fillcolor=bgcolor)

        # affichage separateurs
        self.draw_separator_gantt(rnge)

    def draw_separator_gantt(self, rnge):
        daywidth = self._daywidth
        if self._timestep == 1:
            for n in range(len(rnge)):
                self._handler.draw_line(self._x+n*daywidth, self._y,
                                  self._x+n*daywidth, self._y+ROW_HEIGHT,
                                  color=(204,204,204))
                self._handler.draw_dot(self._x+(n+0.5)*daywidth, self._y,
                                  self._x+(n+0.5)*daywidth, self._y+ROW_HEIGHT,
                                  4,
                                  color=(204,204,204))
        elif self._timestep == 7:
            for n,day in enumerate(rnge):
                if day.day_of_week == 0:
                    self._handler.draw_line(self._x+n*daywidth, self._y,
                                       self._x+n*daywidth, self._y+ROW_HEIGHT,
                                       color=(204,204,204))
                else:
                    self._handler.draw_dot(self._x+n*daywidth, self._y,
                                       self._x+n*daywidth, self._y+ROW_HEIGHT,
                                       4,
                                       color=(204,204,204))          

        else: # timestep == month
            for n,day in enumerate(rnge):
                if day.day == 1:
                    self._handler.draw_line(self._x+n*daywidth, self._y,
                                      self._x+n*daywidth, self._y+ROW_HEIGHT,
                                      color=(204,204,204))
                #elif day.day_of_week == 0:
                #    self._handler.draw_dot(self._x+n*daywidth, self._y,
                #                       self._x+n*daywidth, self._y+ROW_HEIGHT,
                #                       4,
                #                       color=(204,204,204))          
                # les pointilles genent la lecture du graphe

    def _task_timeline(self, worked, is_container, first, last, begin, end, day, factor):
        """
        write a day for a task
        """
        width = self._daywidth / factor
        if not worked:
            self._x += width
            return

        OFFSETS = {
            (True,True)  : (0, 0, 0, 0),
            (True,False) : (1, 1, -1, -2),
            (False,True) : (0, 1, -1, -2),
            (False,False): (0, 1, 0, -2), }
        dx, dy, dw, dh = OFFSETS[ (begin,end) ]

        color = self._color
        coords = self._tasks_slots.setdefault(self._ctask, [])
        coords.append( (self._x, self._y) )

        # draw bg and fg rectangles
        if worked and not is_container:
            self._handler.draw_rect(self._x+dx,
                                    self._y+dy+ROW_HEIGHT*0.125,
                                    max(width+dw, 0),
                                    ROW_HEIGHT*0.75+dh, fillcolor=color)
        # draw... what?
        elif worked and is_container:
            x = self._x
            line_width = round(ROW_HEIGHT/12.)
            y = self._y+5*line_width
            end_width = ROW_HEIGHT/4
            r_x = x
            r_width = width
            self._handler.draw_rect(r_x, y, max(r_width, 0),
                           ROW_HEIGHT-10*line_width, fillcolor=color)

            if first:
                self._handler.draw_poly(((x, y),
                                         (x+end_width, y),
                                         (x, y+ROW_HEIGHT*7/12)),
                                        fillcolor=color)
                r_x = r_x +5
                r_width -= 5
            if last:
                x = x + width
                self._handler.draw_poly(((x, y),
                                         (x-end_width, y),
                                         (x, y+ROW_HEIGHT*7/12)),
                                        fillcolor=color)
                r_width -= 5
        self._x += width

    def milestone_timeline(self, day, milestone, project):
        """
        Iterate over each day to draw corresponding milestone
        """
        self._ctask = milestone
        last_day = day + self._timestep
        begin, end = project.get_task_date_range(milestone)
        assert begin == end
        for day in date_range(day, last_day):
            draw = (day == begin)
            self._milestone_timeline(day, draw, project.factor)

    def _milestone_timeline(self, day, draw, factor):
        """
        Effectively draw a milestone
        """
        # background color
        if day.date == TODAY.date :
            bgcolor = self._color_set['TODAY']
        elif day.day_of_week in (5, 6):
            bgcolor = self._color_set['WEEKEND']
        else:
            bgcolor = self._color_set['WEEKDAY']

        width = self._daywidth
        first_day = self._timeline_days[0]
        last_day = self._timeline_days[-1]+self._timestep
        rnge = list( date_range( first_day, last_day ) )
        self._handler.draw_rect(self._x, self._y, max(width, 0),
                          ROW_HEIGHT, fillcolor=bgcolor)
        self.draw_separator_gantt([day])

        # draw milestone as diamond
        if draw:
            x, y = self._x, self._y
            self._tasks_slots.setdefault(self._ctask, []).append((x, y))
            self._handler.draw_poly(((x+(width-2)/factor, y+ROW_HEIGHT/2), #coin droit
                                     (x+width/(2*factor), y+ROW_HEIGHT*3/4), #haut
                                     (x+2, y+ROW_HEIGHT/2), #gauche
                                     (x+width/(2*factor), y+ROW_HEIGHT/4)), # bas
                                    fillcolor=self._colors['CONSTRAINT'])
        # record position
        self._x += width
        
    def task_constraints(self, type_constraint, task, constraint_task, factor):
        """
        draw a constraint between from task to constraint_task
        """
        # check that constrained task is in the diagram
        if not self._tasks_slots.has_key(constraint_task) or \
               not self._tasks_slots.has_key(task):
            return
        if type_constraint.startswith('begin'):
            index1 = 0
            offset1 = 0
        else:
            index1 = -1
            offset1 = self._daywidth
        if type_constraint.endswith('begin'):
            index2 = 0
            offset2 = 0
        else:
            index2 = -1
            offset2 = self._daywidth / factor
        x1, y1 = self._tasks_slots[task][index1]
        x1 += offset1
        y1 += ROW_HEIGHT/2
        x2, y2 = self._tasks_slots[constraint_task][index2]
        x2 += offset2
        y2 += ROW_HEIGHT/2
        # split line according to differents configuration
        # just for a better visibility
        if x1 > x2:
            x_ = (x1+x2) / 2
            self._handler.draw_line(x1, y1, x_, y1,
                                    color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x_, y2, x2, y2,
                                    color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x_, y1, x_, y2,
                                    color=self._colors['CONSTRAINT'])
        elif y2 <= y1:
            self._handler.draw_line(x2, y2,
                              x2+FIELD_COLUMN_WIDTH/3, y2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x2+FIELD_COLUMN_WIDTH/3, y2,
                              x2+FIELD_COLUMN_WIDTH/3, y1-ROW_HEIGHT/2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x2+FIELD_COLUMN_WIDTH/3, y1-ROW_HEIGHT/2,
                              x1, y1-ROW_HEIGHT/2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x1, y1-ROW_HEIGHT/2,
                              x2-FIELD_COLUMN_WIDTH/3, y1-ROW_HEIGHT/2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x1, y1-ROW_HEIGHT/2,
                              x1-FIELD_COLUMN_WIDTH/3, y1-ROW_HEIGHT/2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x1-FIELD_COLUMN_WIDTH/3, y1-ROW_HEIGHT/2,
                              x1-FIELD_COLUMN_WIDTH/3, y1,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x1-FIELD_COLUMN_WIDTH/3, y1,
                              x1, y1,
                              color=self._colors['CONSTRAINT'])
        else:
            self._handler.draw_line(x2, y2,
                              x2+FIELD_COLUMN_WIDTH/3, y2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x2+FIELD_COLUMN_WIDTH/3, y2,
                              x2+FIELD_COLUMN_WIDTH/3, y1+ROW_HEIGHT/2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x2+FIELD_COLUMN_WIDTH/3, y1+ROW_HEIGHT/2,
                              x1, y1+ROW_HEIGHT/2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x1, y1+ROW_HEIGHT/2,
                              x2-FIELD_COLUMN_WIDTH/3, y1+ROW_HEIGHT/2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x1, y1+ROW_HEIGHT/2,
                              x1-FIELD_COLUMN_WIDTH/3, y1+ROW_HEIGHT/2,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x1-FIELD_COLUMN_WIDTH/3, y1+ROW_HEIGHT/2,
                              x1-FIELD_COLUMN_WIDTH/3, y1,
                              color=self._colors['CONSTRAINT'])
            self._handler.draw_line(x1-FIELD_COLUMN_WIDTH/3, y1,
                              x1, y1,
                              color=self._colors['CONSTRAINT'])
            
        self._handler.draw_poly(((x1+2, y1), (x1-4, y1+4), (x1-4, y1-4)),
                          fillcolor=self._colors['CONSTRAINT'])



"""
Projman - (c)2000-2006 LOGILAB <contact@logilab.fr> - All rights reserved.

Home: http://www.logilab.org/projects/projman

This code is released under the GNU Public Licence v2. See www.gnu.org.

"""

from pprint import pprint
import sys
import os.path as osp

from logilab.common.testlib import unittest_main, TestCase
from logilab.common.compat import set
from projman.lib import *
from projman.scheduling import schedule
from projman.scheduling.simple import SimpleScheduler, cmp_tasks
from projman.scheduling.csp import *
from projman.readers import ProjectXMLReader
from mx.DateTime import DateTime,  Time

def print_solutions(solutions):
    for s in solutions:
        pprint(s)

DATADIR = osp.abspath(osp.join(osp.dirname(__file__), "data"))

class RawSchedulingTC(TestCase):

    def skip_test_solve_simple(self):
        # 2 tasks, no time -> no solution        
        variables = []
        domains = {}
        constraints = []
        for i in range(2):
            name = 't%02d' % i
            variables.append(name)
            domains[name] = SchedulingDomain(0, 2, 2)

        constraints.append(fi.NoOverlap('t00', 't01'))

        repo = Repository(variables, domains, constraints)
        solver = Solver(SchedulingDistributor())
        solutions = solver.solve(repo, verbose=0)
        #print_solutions(solutions)
        self.assertEquals(len(solutions), 0)

    def skip_test_solve_harder(self):
        variables = []
        domains = {}
        constraints = []
        for i in range(2):
            name = 't%02d' % i
            variables.append(name)
            domains[name] = SchedulingDomain(0, 4, 2)

        constraints.append(fi.NoOverlap('t00', 't01'))
        constraints.append(fi.StartsAfterEnd('t01', 't00'))

        repo = Repository(variables, domains, constraints)
        solver = Solver(SchedulingDistributor())
        solutions = solver.solve(repo, verbose=0)
        #print_solutions(solutions)
        self.assertEquals(len(solutions), 1)

    def skip_test_solve_harder2(self):
        variables = []
        domains = {}
        constraints = []
        for i in range(2):
            name = 't%02d' % i
            variables.append(name)
            domains[name] = SchedulingDomain(0, 4, 2)
        variables.append("t02")
        domains['t02'] = SchedulingDomain(0, 6, 2)
        constraints.append(fi.NoOverlap('t00', 't01'))
        constraints.append(fi.NoOverlap('t00', 't02'))
        constraints.append(fi.NoOverlap('t02', 't01'))
        constraints.append(fi.StartsAfterEnd('t02', 't00'))
        constraints.append(fi.StartsAfterEnd('t02', 't01'))

        repo = Repository(variables, domains, constraints)
        solver = Solver(SchedulingDistributor())
        solutions = solver.solve(repo, verbose=0)
        #print_solutions(solutions)
        self.assertEquals(len(solutions), 2)



class CSPSchedulerTC(TestCase):

    def setUp(self):
        filename = osp.join(DATADIR, "csp_scheduling_projman.xml")
        reader = ProjectXMLReader(filename)
        self.project, files = reader.read()

        
    def skip_test_visit(self):
        scheduler = CSPScheduler(self.project)
        scheduler.schedule()
        self.assertEquals(len(scheduler.variables), 4)
        self.assertEquals(len(scheduler.domains), 4)
        expected_constraints = set([
            fi.StartsAfterEnd(u't2_2', u't2_11'),
            fi.StartsAfterEnd(u't2_2', u't2_12'),
            fi.StartsAfterEnd(u't2_3', u't2_2'),
            fi.NoOverlap(u't2_2', u't2_12'),
            ])
        self.assertEquals(expected_constraints, scheduler.constraints)


class CmpTasksTC(TestCase):
        
    def skip_test_cmp(self):
        t0 = Task('0')
        t1 = Task('1')
        self.assertEquals( cmp_tasks(t0,None),  1)
        t1.add_task_constraint('begin-after-end', '0')
        self.assertEquals( cmp_tasks(t0,t1),  -1)

    def skip_test_cmp2(self):
        t0 = Task('0')
        t1 = Task('1')
        self.assertEquals( cmp_tasks(t0,None),  1)
        t0.priority = 2
        self.assertEquals( cmp_tasks(t0,t1),  -1)
        t1.priority = 1
        self.assertEquals( cmp_tasks(t0,t1),  1)
        t0.priority = 1
        self.assertEquals( cmp_tasks(t0,t1),  0)

        
class SimpleSchedulerTC(TestCase):

    def setUp(self):
        filename = osp.join(DATADIR, "csp_scheduling_projman.xml")
        reader = ProjectXMLReader(filename)
        self.project, files = reader.read()
        
    def skip_test_visit1(self):
        scheduler = SimpleScheduler(self.project)
        ordered_buckets = scheduler.get_ordered_buckets()
        self.assertEquals(len(ordered_buckets), 3)
        self.assertEquals(len(ordered_buckets[0]), 2)
        self.assertEquals(len(ordered_buckets[1]), 1)
        self.assertEquals(len(ordered_buckets[2]), 1)

    def skip_test_visit2(self):
        scheduler = SimpleScheduler(self.project)
        scheduler.schedule()

if __name__ == '__main__':
    unittest_main()



"""
Projman - (c)2000-2007 LOGILAB <contact@logilab.fr> - All rights reserved.

Home: http://www.logilab.org/projman

Manipulate a xml project description.

This code is released under the GNU Public Licence v2. See www.gnu.org.
"""

from mx.DateTime import DateTime, Time, Date
from logilab.common import testlib
from projman.lib import Project, Calendar, ResourcesSet, Resource

MORNING = (Time(8), Time(12))
AFTERNOON = (Time(13), Time(17))
HALFDAY = (Time(9), Time(15))

def mk_calendar1():
    cal = Calendar('c_1', 'Calendrier 1')
    cal.day_types = {'nonworking': ['Non working', []],
                     'working':    ['Working', [MORNING, AFTERNOON]],
                     
                         'halfday':    ['HalfDay', [HALFDAY]],
                         }
    cal.default_day_type = 'working'
    cal.national_days = [(1,1), (12,25), (11,11)]
    cal.start_on = DateTime(2004,01,06) 
    cal.stop_on = DateTime(2006,12,29)
    cal.weekday['mon'] = 'working'
    cal.weekday['tue'] = 'working'
    cal.weekday['wed'] = 'working'
    cal.weekday['thu'] = 'working'
    cal.weekday['fri'] = 'working'
    cal.weekday['sat'] = 'nonworking'
    cal.weekday['sun'] = 'nonworking'

    date = DateTime(2004, 05, 07)
    cal.add_timeperiod(date, date, 'nonworking')
    date = DateTime(2004, 05, 05)
    cal.add_timeperiod(date, date, 'working')
    date = DateTime(2004, 06, 07)
    cal.add_timeperiod(date, date, 'halfday')
    return cal

def mk_calendar2():
    cal = Calendar('c_2', 'Calendrier 2') 
    from_date = DateTime(2004, 06, 12)
    to_date = DateTime(2004, 06, 23)
    cal.add_timeperiod(from_date, to_date, 'nonworking')
    return cal

class CalendarTC(testlib.TestCase):
    """
    Calendar represents
    """
    def setUp(self):
        """ called before each test from this class """
        self.o = Project()
        self.o.title = 'Projman'

        self.rss = ResourcesSet('all_resources')

        self.r1 = Resource('r_1', 'Resource 1')
        self.r1.calendar = 'c_1'
        self.r1.type = '1'
        self.rss.add_resource(self.r1)
        
        self.r2 = Resource('r_2', 'Resource 2')
        self.r2.calendar = 'c_2'
        self.r2.type = '1'
        self.rss.add_resource(self.r2)
        
        self.c1 = mk_calendar1()
        self.c2 = mk_calendar2()
        self.c1.append(self.c2)
        self.rss.add_calendar(self.c1)

        self.o.add_resource_set(self.rss)

    def test_within_bounds(self):
        self.assertEquals(self.c1.after_start(Date(2003,1,1)), False)
        self.assertEquals(self.c1.after_start(Date(2004,1,5)), False)
        self.assertEquals(self.c1.after_start(Date(2004,1,6)), True)
        self.assertEquals(self.c1.after_start(Date(2005,1,1)), True)
        self.assertEquals(self.c1.before_stop(Date(2003,1,1)), True)
        self.assertEquals(self.c1.before_stop(Date(2006,12,30)), False)

    def test_availability(self):
        DAY = [MORNING, AFTERNOON]
        self.assertEqual(self.c1.availability(DateTime(2004, 01, 05)), [])
        self.assertEqual(self.c1.availability(DateTime(2004, 01, 06)), DAY)
        self.assertEqual(self.c1.availability(DateTime(2004, 05, 07)), [])
        self.assertEqual(self.c1.availability(DateTime(2004, 06, 07)), [HALFDAY])
        self.assertEqual(self.c1.availability(DateTime(2006, 12, 29)), DAY)
        self.assertEqual(self.c1.availability(DateTime(2006, 12, 30)), [])
        self.assertEqual(self.c2.availability(DateTime(2004, 05, 07)), [])

    def test_get_total_intervals(self):
        self.assertEqual(self.c1.get_worktime('nonworking').hours, 0)
        self.assertEqual(self.c1.get_worktime('working').hours, 8)
        self.assertEqual(self.c2.get_worktime('halfday').hours, 6)

    def test_is_a_national_day(self):
        self.assertEqual(self.c1.is_a_national_day(DateTime(2012, 01, 01)), True)
        self.assertEqual(self.c1.is_a_national_day(DateTime(2012, 04, 01)), False)
        self.assertEqual(self.c2.is_a_national_day(DateTime(2012, 01, 01)), True)
        self.assertEqual(self.c2.is_a_national_day(DateTime(2012, 02, 02)), False)

    def test_get_daytype(self):
        self.assertEquals(self.c1.get_default_daytype(), 'working')
        self.assertEquals(self.c2.get_default_daytype(), 'working')
        
if __name__ == '__main__':
    testlib.unittest_main()



# Copyright (c) 2000-2002 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
reader generate a model from xml file (see dtd/project.dtd)
"""
__revision__ = "$Id: base_reader.py,v 1.2 2005-11-09 16:47:35 arthur Exp $"

from os.path import dirname, abspath
from xml.sax import make_parser, ContentHandler
from xml.sax.handler import feature_namespaces

from projman.lib.project import Project
from projman.lib.task import Task, MileStone
from projman.lib.resource import Resource, ResourcesSet
from projman.lib.resource_role import ResourceRole, ResourceRoleSet
from projman.lib.calendar import Calendar
from projman.lib._exceptions import ProjectValidationError, MalformedProjectFile


class ModelProjectFactory:
    """ a factory which create Projman model's objects """
    
    def create_project(self):
        """ create a new projman node """
        return Project()
        
    def create_milestone(self, m_id):
        """ create a new milestone node """
        return MileStone(m_id)
        
    def create_milestone_from_task(self, task):
        """ create a new milestone node """
        stone = self.create_milestone(task.id)
        stone.title = task.title
        stone.task_constraints = task.task_constraints
        stone.description = task.description
        return stone

    def create_task(self, t_id):
        """ create a new task node """
        return Task(t_id)

    def create_resource(self, r_id, r_name, r_type, calendar, r_hourly_rate='0.0', r_unit='euros'):
        """ create a new resource """
        try:
            float_rate = float(r_hourly_rate)
        except ValueError:
            raise AttributeError("Resource's hourly rate must be a float")
        return Resource(r_id, r_name, r_type, calendar, float_rate, r_unit)

    def create_resourcesset(self, rs_id):
        """ create a new resources set """
        return ResourcesSet(rs_id)

    def create_resource_role(self, r_id, name, cost='0.0', unit='euros'):
        """create a new resource_role"""
        return ResourceRole(r_id, name, cost, unit)

    def create_resource_role_set(self, rs_id):
        """ create a new resource_role set """
        return ResourceRoleSet(rs_id)

    def create_calendar(self, tt_id, tt_name = u''):
        """ create a new time table """
        return Calendar(tt_id, tt_name)

    def subproject_to_task(self, project):
        """ converts a subproject to a task """
        t = self.create_task(project.id)
        for task in project.children:
            t.append(task)
        return t

MODEL_FACTORY = ModelProjectFactory()

# abstract reader for xml files ###############################################

class AbstractXMLReader(ContentHandler):
    """ generic abstract reader """
    
    def __init__(self, factory=MODEL_FACTORY):
        self._factory = factory
        # default type used for resources list
        self._default_type = ""
        # path in the tree
        self.stack = []
        self._tags = []
        # base uri used for imported file resolution
        self._base_uris = []
        # file names for error report
        self._files = []
        # keep trace of yet imported files
        self._imported = {}
        # buffer for characters() callback
        self._buffer = []
        # buffer of formatting errors
        self._errors = []
       
    def fromFile(self, file):
        """ 
        import and return a project from a xml file
        """
        return self.fromStream(open(file),
                               file, dirname(abspath(file)))
        
    def fromStream(self, stream,
                   filename="input_stream", base_uri=''):
        """ 
        import and return a project from a xml stream
        """
        # create a xml parser
        p = make_parser()
        # do not perform Namespace processing 
        p.setFeature(feature_namespaces, 0)
        p.setContentHandler(self)
        p.reset()

        # parse file
        self._base_uris.append(base_uri)
        self._files.append(filename)
        p.parse(stream)
        self._base_uris.pop()
        self._files.pop()
        if self._errors:
            raise MalformedProjectFile('\n'.join(self._errors))
        return self._custom_return()

    def _custom_return(self):
        raise NotImplementedError(self.__class__)

    def startDocument(self):
        self._buffer = []

    def characters(self, data):
        self._buffer.append(data)

    def get_characters(self, joined_with=' ', strip=True):
        """return the character buffer.
        By default, strips all chunks and join them with a single space.
        If strip is false, join is performed with an empty string."""
        if strip:
            self._buffer = [a.strip() for a in self._buffer if a.strip()]
        else:
            joined_with = u''
        characters = joined_with.join(self._buffer)
        self._buffer = []
        return characters

    def startElement(self, tag, attr):
        try:
            self._attrs = attr
            self._start_element(tag, attr)
        except ProjectValidationError, error:
            try:
                msg = unicode(error)% (self._files[-1],
                                       self._locator.getLineNumber(),
                                       tag)
                self._errors.append(msg)
                self._tags.append(tag)
            except TypeError:
                raise error
        except:
            if self._errors:
                print 'WARNING: pending validation errors:'
                print '\n'.join(self._errors)
            raise
        self._tags.append(tag)

    def _start_element(self, tag, attr):
        pass
    
    def endElement(self, tag):
        try:
            self._end_element(tag)
        except ProjectValidationError, error:
            try:
                msg = unicode(error)% (self._files[-1],
                                       self._locator.getLineNumber(),
                                       tag)
                self._errors.append(msg)
            except TypeError:
                raise error
        except:
            if self._errors:
                print 'WARNING: pending validation errors:'
                print '\n'.join(self._errors)
            raise
        self._tags.pop()
        

    def _end_element(self, tag):
        pass

    def assert_child_of(self, tags):
        if not self._tags :
            if None in tags :
                return
            else:
                raise ProjectValidationError('file %s line %s : %s is not root tag')
        elif not self._tags[-1] in tags:
            raise ProjectValidationError('file %%s line %%s : %%s should be child of '
                                         '%s not %s' % (tags, self._tags[-1]))
        
    def assert_has_attrs(self, attrs):
        for attr in attrs:
            if attr not in self._attrs.keys():
                # TODO replace by MissingRequiredAttribute
                raise ProjectValidationError('file %%s line %%s : Missing required ' \
                                             'attribute %s to element <%%s>' % attr)
                #msg = unicode(error)% (self._files[-1],
                #                       self._locator.getLineNumber(),
                #                       tag)
                #self._errors.append(msg)




# -*- coding: iso-8859-1 -*-
"""
unit tests for module logilab.projman.lib.Resource

Projman - (c)2000-2002 LOGILAB <contact@logilab.fr> - All rights reserved.

Home: http://www.logilab.org/projman

Manipulate a xml project description.

This code is released under the GNU Public Licence v2. See www.gnu.org.

"""
__revision__ = "$Id: unittest_lib_Resource.py,v 1.11 2005-09-06 18:27:45 nico Exp $"

from logilab.common.testlib import TestCase, unittest_main

from projman.lib import Resource, Calendar, ResourcesSet
from mx.DateTime import DateTime, Time
    
class ResourceTest(TestCase):
    """
    Resource represents 
    """
    def setUp(self):
        """ called before each test from this class """
        # set up resources
        self.r1 = Resource('r_1', 'Resource 1')
        self.r1.calendar = 'c_1'
        self.r2 = Resource(u'où', u'là-bas')
        self.r2.calendar = 'c_2'
        # set of dates
        self.date_last_week = DateTime(2004, 10, 1)
        self.date_today = DateTime(2004, 10, 7)
        self.date_tomorrow = DateTime(2004, 10, 8)
        self.date_next_week = DateTime(2004, 10, 13)
        self.date_next_score = DateTime(2004, 10, 26)
        # set up calendar 1      
        self.c1 = Calendar('c_1', 'Defaut')
        self.c1.day_types = {'working':['Working', [(Time(8), Time(12)),
                                                    (Time(13), Time(17))]],
                             'halfday':['HalfDay', [(Time(9), Time(15))]],
                             'nonworking': ['Nonworking', []],
                             } 
        self.c1.default_day_type = 'working'
        self.c1.add_timeperiod(self.date_last_week, self.date_last_week, 'nonworking')
        self.c1.add_timeperiod(self.date_today, self.date_today, 'working')
        self.c1.add_timeperiod(self.date_tomorrow, self.date_next_week, 'halfday')
        self.c1.weekday['sat'] = 'nonworking'
        self.c1.weekday['sun'] = 'nonworking'
        # set up calendar 2        
        self.c2 = Calendar('c_2', u'Année 2')
        self.c2.add_timeperiod(self.date_next_week, self.date_next_score, 'nonworking')
        # build tree
        self.c1.append(self.c2)
        self.rss = ResourcesSet('all_resources')
        self.rss.add_resource(self.r1)
        self.rss.add_resource(self.r2)
        self.rss.add_calendar(self.c1)
        
    def test_get_default_wt_in_hours(self):
        self.assertEqual(self.r1.get_default_wt_in_hours(), 8)
        
    def test_is_available(self):
        """
        tests if a resource is available on datetime according to its calendar
        """
        # test r_1
        self.assertEqual(self.r1.is_available(self.date_last_week), False)
        self.assertEqual(self.r1.is_available(self.date_today), True)
        self.assertEqual(self.r1.is_available(self.date_tomorrow), True)
        self.assertEqual(self.r1.is_available(self.date_next_week), True)
        # test r_2
        self.assertEqual(self.r2.is_available(self.date_last_week), False)
        self.assertEqual(self.r2.is_available(self.date_today), True)
        self.assertEqual(self.r2.is_available(self.date_tomorrow), True)
        self.assertEqual(self.r2.is_available(self.date_next_week), False)
        self.assertEqual(self.r2.is_available(self.date_next_score), False)
        
    def test_get_duration_of_work(self):
        """
        return the total number of seconds of work at datetime
        """
        self.assertEqual(self.r1.get_worktime(self.date_next_week).hours, 6)
        self.assertEqual(self.r2.get_worktime(self.date_next_week).hours, 0)
        self.assertEqual(self.r2.get_worktime(self.date_tomorrow).hours, 6)
        self.assertEqual(self.r2.get_worktime(self.date_today).hours, 8)

if __name__ == '__main__':
    unittest_main()



# Copyright (c) 2000-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
base scheduling visitors
"""

__revision__ = "$Id: basic.py,v 1.2 2005-09-08 14:26:06 nico Exp $"

import sys

from logilab.common.visitor import Visitor
from logilab.common.tree import PostfixedDepthFirstIterator

from projman.lib import date_range
from projman.lib.constants import *
from mx.DateTime import now, DateTimeFromAbsDateTime
                

def add_week_days(date, days):    
    """
    adds a number of days taking into account weekends
    """
    #short circuit - taking into consideration weekends doesn't seem necessary
    return date + days
##     day_of_week = date.day_of_week
##     days_left = 4 - day_of_week
##     if days <= days_left:
##         return date+days
##     weeks = days // 5
##     days_left = days % 5
##     date += 7 * weeks
##     if days_left:
##         if date.day_of_week + days_left >= 5:
##             date += 2 + days_left
##         else:
##             date += days_left
##     return date

def sub_week_days(date, days):
    # quick and dirty since it is not used often
    weeks = days // 5
    remain = days % 5
    return date - (7 * weeks) - remain

def split_to_fit_disponibility(activity, resource, a_seq=None):
    """
    Takes an activity and returns a sequence of activities that 
    take into account the days the resource cannot work
    """
    if a_seq is None:
        a_seq = []
    begin = activity.begin
    date = begin
    end = activity.end
    while date <= end :
        if resource.is_available(date): 
            date += 1
        else: 
            if date != activity.begin:
                a_seq.append( (begin, date-1, activity.usage) )
            begin = date
            end += 1
    a_seq.append( (begin, date-1, activity.usage) )
    return a_seq

# Basic Scheduling visitor ####################################################

class BasicScheduler(Visitor):
    """
    The first step schedule visitor
    
    get the ASAP begin and end date for each task of the tree
    """
    
    def __init__(self, project):
        """
        create a BasicScheduler. Use open_visit() to start.
        """
        Visitor.__init__(self, PostfixedDepthFirstIterator)
        self._root = None
        self.errors = []
        self.project = project
        self._past_activities = Table()
        self._current_schedule = None

    def schedule(self, verbose=1, **kw):
        """
        Try to update the current_schedule using data from project
        and activities.
        
        Return project_errors
        """
        self.errors = []
        self._schedule = self.project.schedule
        return self.visit(self.project.root_task)

    # visitor API #############################################################
    
    def open_visit(self, node):
        self._root = node

    def close_visit(self, result):
        return self.errors
    
    def visit_project(self, node):
        # FIXME - is this really useful?
        if node.parent is None:
            schedule = self.project.get_schedule(node.id) 
            assert schedule is not None
            
        self.visit_task(node)
       
        if node.parent is None:
            begin, end = self.project.get_date_range(node)
            schedule.global_begin = begin
            schedule.global_end = end
    
    def visit_milestone(self, node):
        if hasattr(node, 'scheduled'):
            return
        node.scheduled = 1
        
        if node.children:
            raise ValueError , "Milestone can't have children"
        else :
            self.schedule_milestone(node)
   
    def visit_task(self, node):
        if hasattr(node, 'scheduled'):
            return
        node.scheduled = 1        
        if node.children:
            self.schedule_container(node)
        else :
            self.schedule_leaf(node)
        begin, end = self.project.calculate_date_range(node)
        self._current_schedule.tasks_timeslot[node.id] = [begin, end]
        # set also global_tasks_cost and tasks_cost_by_resource
        all_activities = self._current_schedule._merge_activities.select('task', node.id)
        #  compute the global cost of task according to time-slots and resources cost
        global_cost = 0.0
        unit = 'euros'
        for r_id in all_activities:
            r = self.project.get_resource(r_id)
            hours_per_day = r.get_default_wt_in_hours()
            list_activity = all_activities[r_id]
            hours = 0.0
            nb_days = 0.0
            for activity in list_activity:
                d = activity.begin
                loop_count = 0
                while d <= activity.end:
                    worked_hours = (r.get_duration_of_work(d) / 3600.) * activity.usage
                    # FIXME XXX infiniteloop BUG prevention
                    if loop_count == 100:
                        break
                    if worked_hours == 0 and r.is_available(d):
                        break
                    hours += worked_hours
                    d += worked_hours/hours_per_day
                    nb_days += worked_hours/hours_per_day
                    loop_count += 1
            r_hourly_rate = r.hourly_rate[0]
            unit = r.hourly_rate[1]
            global_cost += hours * r_hourly_rate
            # update cost in resource
            task_cost_by_resource = self._current_schedule.tasks_cost_by_r.setdefault(node.id, {})
            task_cost_by_resource[r_id] = task_cost_by_resource.get(r_id, 0) + nb_days
            
        # update cost in schedule
        if node.id not in self._current_schedule.tasks_global_cost:
            if global_cost != 0:
                self._current_schedule.tasks_global_cost[node.id] = [global_cost, unit]
        else:
            #FIXME need currency converter if unit != global_unit
            assert self._current_schedule.tasks_global_cost[node.id][1] == unit, \
                   "Panic: we don't support different units yet"
            self._current_schedule.tasks_global_cost[node.id][0] += global_cost

    # methods for the real job ################################################

    def schedule_container(self, node):
        """
        set begin and end for a container task according to its children tasks
        """
        # simply define begin and end
        begin, end = self.project.get_container_date_range(node, 
                                                   self._current_schedule) 
        if begin and end:
            self._current_schedule.tasks_timeslot[node.id] = [begin, end]
        else:
            sys.stderr.write('warning... unscheduled leaves!\n')
        
        
    def schedule_milestone(self, node) :
        """ 
        get node begin / end date
        set time slots according to estimated begin and end of the task
        """
        for tc_type in TASK_CONSTRAINTS:
            n_cons = self.project.get_task_constraints(node.id,
                                                        fill = TASK_CONSTRAINTS) 
            if tc_type in n_cons:
                for task_id in n_cons[tc_type]:
                    if node.get_node_by_id(task_id) is not None:
                        v = BasicScheduler(self.project)
                        v.visit(node.get_node_by_id(task_id), 
                                self._past_activities,
                                self._current_schedule)
                        self.errors += v.errors
        date, date_strict = self.get_possible_begin(node)
        self._current_schedule.tasks_timeslot[node.id] = [date, date]
        # FIXME is activity really necessary? check usage by renderer
        self._current_schedule.planned_activities.add_activity(node.id, '',
                                                               Activity(100, date, date))
    
    def schedule_leaf(self, node) :
        """ 
        get node begin / end date
        set time slots according to estimated begin and end of the task
        """
        if node.is_finished(self.project.get_activity_by_t(node.id)):
            self._current_schedule.tasks_status[node.id] = 'done'
        else:
            planned_activities = self._current_schedule.planned_activities
            # schedule all constrained task
            for tc_type in TASK_CONSTRAINTS:
                n_cons = self.project.get_task_constraints(node.id,
                                                            fill=TASK_CONSTRAINTS)
                for task_id in n_cons.get(tc_type, []):
                    task = node.get_node_by_id(task_id)
                    if task is not None:
                        v = BasicScheduler(self.project) # !?!
                        v.visit(task, self._past_activities,
                                self._current_schedule)
                        self.errors += v.errors
            # calculate edges
            begin, begin_strict = self.get_possible_begin(node)
            end, end_strict = self.get_possible_end(node)
            nb_days = node.duration
            res_seq = node.get_resources()
            # include duration of past activities in schedule
            nb_days -= self._past_activities.get_duration_worked_for_t(node.id)
            # cut up according to days worked
            date = begin
            end_limit = end_strict and end or end + 365
            # FIXME we end up having a great number of daily activities.
            while date <= end_limit:
                # calculate global disponibility at 'date'
                dispo = {}
                for resource in res_seq: 
                    dispo[resource] = min(
                        self.project.get_general_disponibility(resource, date),
                        node.get_resource_dispo(resource))
                sum_d = sum(dispo.values())
                # dispo sufficient to finish task: spread out activity among resources
                if nb_days - sum_d < -FLOAT_ZERO:
                    #FIXME TODO : if a resource available less than spread out 
                    #value... then you have more on the others (end up needing
                    #an extra day etc...)
                    used_resources = dispo.keys()
                    ideal_usage = nb_days*1./len(used_resources)
                    for resource in used_resources:
                        real_usage = min(ideal_usage, dispo[resource])
                        planned_activities.add_activity(date, date, resource, node.id, real_usage)
                        nb_days -= real_usage
                # dispo not sufficient: std option, fill in full days
                else:
                    for resource, usage in dispo.items():
                        if usage:
                            planned_activities.add_activity(date, date, resource, node.id, usage)
                            nb_days -= usage
                # task filled with activities, break
                if nb_days <= FLOAT_ZERO:
                    break
                date += 1
            # check that task has been complete (all days in duration linked to an activity)
            if nb_days > FLOAT_ZERO:
                self.errors.append('Unable to fit all days required before end date of %s (%s)' \
                                   % (node.id, nb_days))
                self._current_schedule.tasks_status[node.id] = 'problem'
                node.problem = True
           
    def get_possible_begin(self, node):
        """
        return the possible begin date of the task
        
        check for constraints inconsistency, remove broken constraints
        """
        pl_act = self._current_schedule.planned_activities 
        begin = None
        date_cs = node.get_date_constraints(fill = DATE_CONSTRAINTS)
        # set this variable to 1 if the begin date can not be later
        strict = None
        TODAY = now()

        # check begin-at / end-at date constraints
        if date_cs[BEGIN_AT_DATE] is not None:
            # remove end-at-date constraint if any
            if date_cs[END_AT_DATE] is not None:
                date_cs[END_AT_DATE] = None
            date = date_cs[BEGIN_AT_DATE]
            if begin is not None and date != begin  or date < TODAY:
                self.remove_constraint(node, BEGIN_AT_DATE,
                                       date_cs[BEGIN_AT_DATE])
                date_cs[BEGIN_AT_DATE] = None
            else:
                strict = 1
                begin = date
        elif date_cs[END_AT_DATE]:
            date = date_cs[END_AT_DATE]
            if node.TYPE == 'milestone':
                begin_est = date
            else : 
                begin_est = sub_week_days(date, \
                        node.remaining_duration()) #self.project))
            if begin is not None and not begin_est == begin or begin < TODAY:
                self.remove_constraint(node, END_AT_DATE,
                                       date_cs[END_AT_DATE])
                date_cs[END_AT_DATE] = None
            else:
                strict = 1
                begin = begin_est

        if begin is None:
            begin = TODAY
                
        # check other begin-date constraints
        if date_cs[BEGIN_AFTER_DATE] is not None:
            if begin < date_cs[BEGIN_AFTER_DATE] and strict is not None:
                self.remove_constraint(node, BEGIN_AFTER_DATE,
                                       date_cs[BEGIN_AFTER_DATE])
                date_cs[BEGIN_AFTER_DATE] = None
            else:
                begin = date_cs[BEGIN_AFTER_DATE]

        if date_cs[BEGIN_BEFORE_DATE] is not None and strict is not None:
            if begin > date_cs[BEGIN_BEFORE_DATE]:
                self.remove_constraint(node, BEGIN_BEFORE_DATE,
                                       date_cs[BEGIN_BEFORE_DATE])
                date_cs[BEGIN_BEFORE_DATE] = None
            else:
                strict = 1

        # check begin-after-end task constraints
        node_cons = self.project.get_task_constraints(node.id, fill = TASK_CONSTRAINTS)
        #self.logger.debug(" -- [%s] task constraints=%s"% (node.id, str(node_cons)))
        new = []
        for task_id in node_cons[BEGIN_AFTER_END]:
            task = node.get_node_by_id(task_id) 
            if task is not None:
                date_end = self.adjust(pl_act.get_end(task), task) or now()
                task_end = date_end
                if begin <= task_end:
                    if strict is not None:
                        self.remove_constraint(node, BEGIN_AFTER_END, task_id)
                    else:
                        new.append(task_id)
                        begin = task_end 
                else:
                    new.append(task_id)
            else:
                pass
        node_cons[BEGIN_AFTER_END] = new
        #XXX TODO:  BEGIN_BEFORE_DATE ???

        # check begin-after-begin task constraints
        new = []
        for task_id in node_cons[BEGIN_AFTER_BEGIN]:
            if node.get_node_by_id(task_id) is not None:
                task = node.get_node_by_id(task_id)
                date_begin = pl_act.get_begin(task) or now() 
                task_begin = date_begin
                if begin < task_begin:
                    if strict is not None:
                        self.remove_constraint(node, 
                                               BEGIN_AFTER_BEGIN, 
                                               task_id)
                    else:
                        new.append(task_id)
                        begin = (DateTimeFromAbsDateTime(task_begin) \
                                  + 1)
                else:
                    new.append(task_id)
            else:
                pass
        node_cons[BEGIN_AFTER_BEGIN] = new
        begin = DateTimeFromAbsDateTime(begin)
        node.scheduled = begin
        return begin, strict

    def get_possible_end(self, node):
        """
        return the estimated end of the task
        
        check for constraints inconsistency, remove broken constraints
        you must call get_possible_begin before get_possible_end
        """
        assert node.scheduled
        pl_act = self._current_schedule.planned_activities 
        date_cs = node.get_date_constraints(fill = DATE_CONSTRAINTS)
        end = node.scheduled + max(0, node.remaining_duration()) #self.project))
        # set this variable to 1 if the begin date can not be latter
        strict = None
            
        # check date constraints
        if date_cs[END_AFTER_DATE] is not None:
            if end < date_cs[END_AFTER_DATE]:
                self.remove_constraint(node, END_AFTER_DATE,
                                       date_cs[END_AFTER_DATE])
                date_cs[END_AFTER_DATE] = None
            else:
                strict = 1
                end = date_cs[END_AFTER_DATE]

        if date_cs[END_BEFORE_DATE] is not None and \
               end > date_cs[END_BEFORE_DATE]:
            self.remove_constraint(node, END_BEFORE_DATE,
                                   date_cs[END_BEFORE_DATE])
            date_cs[END_BEFORE_DATE] = None
        #XXX TODO:  END_AT_DATE ???
                    
        # check task constraints
        node_cons = self.project.get_task_constraints(node.id, 
                                            fill = TASK_CONSTRAINTS)
        new = []
        for task_id in node_cons[END_AFTER_END]:
            task = node.get_node_by_id(task_id) 
            if task is not None:
                date_end = pl_act.get_end(task) or now()
                task_end = date_end
                if end < task_end:
                    if strict is not None:
                        self.remove_constraint(node, END_AFTER_END, task_id)
                    else:
                        new.append(task_id)
                        end = task_end
                else:
                    new.append(task_id)
        node_cons[END_AFTER_END] = new
        
        new = []
        for task_id in node_cons[END_AFTER_BEGIN]:
            if node.get_node_by_id(task_id) is not None:
                date_begin = pl_act.get_begin(task) or now()
                task_begin = date_begin
                if end < task_begin:
                    if strict is not None:
                        self.remove_constraint(node, END_AFTER_BEGIN, task_id)
                    else:
                        new.append(task_id)
                        end = task_begin
                else:
                    new.append(task_id)
        node_cons[END_AFTER_BEGIN] = new
        return end, strict
            
    def remove_constraint(self, node, c_type, c_value):
        """
        notify that an unsatisfable constraint has been removed
        """
        self.errors.append('remove %s %s constraint on %s' % (
            c_type, c_value, node.id))
        self._current_schedule.tasks_status[node.id] = 'problem'
        node.problem = True
    
    def adjust(self, date, task):
        """
        checks preceding days if resources are fully used, if not 
        return previous day
        """
        # define result var in order to ease logging
        result = date or None
        # no adjusting can be done with a dependency on a milestone
        if task.TYPE != 'milestone':
            for res_id in task.get_resources():
                dispo = self.project.get_general_disponibility(res_id, date - 1)
                usage = task.get_resource_dispo(res_id)
                if dispo - usage > 0:
                    result = date - 1
        return result




# Copyright (c) 2000-2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""library that enables the scheduling of a project"""

__revision__ = "$Id: __init__.py,v 1.22 2005-09-07 23:51:01 nico Exp $"

import sys
from logging import getLogger

log = getLogger("schedule")

class ScheduleException(Exception):
    """base of scheduling exceptions"""

class ScheduleError(ScheduleException):
    """scheduling error"""

class NoSolutionFound(ScheduleException):
    """unable to find a solution"""


def schedule(proj, config):
    """
    launch scheduling of a projman instance
    Uses CSPScheduler if csp is true, otherwise BasicScheduler
    """
    errors = []
    proj.reset_schedule()
    _type = config.type
    if _type == 'csp':
        from projman.scheduling.csp import CSPScheduler
        scheduler = CSPScheduler(proj)
    elif _type == 'dumb':
        from projman.scheduling.dumb import DumbScheduler
        scheduler = DumbScheduler(proj)
    elif _type == 'simple':
        from projman.scheduling.simple import SimpleScheduler
        scheduler = SimpleScheduler(proj)
    else:
        raise ValueError('bad scheduler type %s'%_type)
    errors += scheduler.schedule(verbose=config.verbose, time=config.time)
    for error in errors:
        log.error(str(error))



# -*- coding: utf-8 -*-
# Copyright (c) 2005 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Projman - (c)2000-2005 LOGILAB <contact@logilab.fr> - All rights reserved.

Home: http://www.logilab.org/projman

Manipulate a xml project description.

This code is released under the GNU Public Licence v2. See www.gnu.org.
"""

__revision__ = "$Id: project.py,v 1.6 2005-11-10 11:34:20 arthur Exp $"

import md5
from warnings import warn
from pprint import pprint
from mx.DateTime import today

from logilab.common.compat import set
from logilab.common.table import Table
from logilab.common.tree import NodeNotFound

from projman.lib.constants import *
from projman.lib.constants import HOURS_PER_DAY

class Project:
    """A project is made a hierachy of tasks, a set of calendars and
    resources, and can optionally include a list of activities and a
    list of scheduled activities

    :attr root_task: root of task hierarchy
    :attr constraints: table of constraints (caching information
                       from task hierarchy)
    :attr resource_set: ResourcesSet() instance representing list of
                        resources available for the project
    :attr activities: task information
    :attr schedule: Schedule() instance (planned activities)
    """

    TYPE = 'project'

    def __init__(self):
        self.root_task = None
        self.resource_set = None
        self.resource_role_set = None
        self._is_scheduled = False
        self.activities = Table(default_value=None,
                                col_names=['begin', 'end', 'resource', 'task',
                                           'usage', 'src'])
        # Stores computation results based on data from self.activities
        #  (here, 'cost' represents the global cost for the task)
        self.tasks = Table(default_value=None,
                           col_names=['begin', 'end', 'status', 'cost', 'unit'])
        # Stores computation results based on data from self.activities
        #  (here, 'cost' represents the resource's cost for the task)
        self.costs = Table(default_value=None,
                           col_names=['task', 'resource', 'cost', 'unit'])
        self.milestones = {}
        self.factor = 1
        self.priority = 3
        # traite les contraintes de priorité jusque 3 (ie 1, 2 et 3)
        self.nb_solution = 0

    def get_factor(self):
        """find if we must schedule on day, half  or quarter of day
        and return the appropriate factor
        """
        pass # valeur donnee en option ou egale a 1
#        factor = 1
#        factor_ = 1
#        for leaf in self.root_task.leaves():
#            mod = (leaf.duration % 1 )
#            if mod > 0:
#                if mod >= 0.13:
#                    factor_ = mod
#                if mod > 0.5 and mod -0.5 >= 0.13 and mod<0.83:
#                    if mod >= 0.13:
#                        factor_ = (leaf.duration % 1 ) - 0.5
#            factor = min(factor, factor_)
            
#        dist=abs(factor-1.)
#        _factor = 1.
#        for f in [1., 0.5, 0.25]:
#             d=abs(factor-f)
#             if d<dist:
#                _factor = f
#                dist = d
#        factor = _factor
#        self.factor = int(1 / factor)

    def get_root_task(self):
        return self._root_task

    def set_root_task(self, value):
        assert self._root_task is None
        self._root_task = value

    root_task = property(get_root_task, set_root_task)

    def check_consistency(self):
        errors = []
        errors += self.root_task.check_consistency()
        return errors

    # schedule methods ########################################################

    def add_schedule(self, activities):
        """update activities and cache with new schedule-based information"""
        self._is_scheduled = True
        for begin, end, resource, task, usage in activities:
            self.activities.append_row((begin, end, resource, task,
                                        usage, 'plan'))
        self.update_caches()

    def reset_schedule(self):
        self._is_scheduled = False
        self.activities.remove('src', 'plan')
        self.update_caches()

    def compute_task_status(self, task, begin, end):
        if task.progress == 1:
            status = 'done'
        elif task.progress > 0:
            status = 'current'
        else:
            status = 'todo'
        for c_type, date, priority in task.get_date_constraints():
            if c_type == BEGIN_AFTER_DATE and not date <= begin and \
                            priority  <= self.priority:
                print "task %s should begin after %s" % (task.id, begin)
                status = 'problem'
                break
            elif c_type == END_BEFORE_DATE and not end <= date and \
                            priority  <= self.priority:
                print "task %s should end before %s" % (task.id, begin)
                status = 'problem'
                break
        return status

    def update_caches(self):
        """updates self.tasks and self.costs"""
        # FIXME: Table has no empty() method
        self.tasks = Table(default_value=None,
                           col_names=['begin', 'end', 'status', 'cost', 'unit'])
        self.costs = Table(default_value=None,
                           col_names=['task', 'resource', 'cost', 'unit'])
        for task in self.root_task.flatten():
            try:
                begin, end = self.get_task_date_range(task)
                status = self.compute_task_status(task, begin, end)
                cost = self.get_task_total_cost(task.id, task.duration)
                self.tasks.append_row((begin, end, status, cost, 'XXX'),
                                      row_name=task.id)
            except ValueError, exc:
                #print 'while building cache, ignoring', exc
                pass #print 'x',
            else:
                pass #print '.',
                
        grouped = self.activities.groupby('task', 'resource')
        for task_id, resources in grouped.iteritems():
            for res_id, rows in resources.iteritems():
                cost = sum([row[4] for row in rows])
                self.costs.append_row((task_id, res_id, cost, 'XXX'))

    # resources methods #######################################################

    def add_resource_set(self, res_set):
        warn('ResourceSet.merge is not yet implemented, replacing instead',
             DeprecationWarning, stacklevel=2)
        #self.resource_set.merge(res_set)
        self.resource_set = res_set

    def has_resource(self, res_id):
        """tests if resource <res_id> is used"""
        return res_id in self.get_resources()

    def get_resources(self):
        """return all the resources available for the project"""
        return self.resource_set.get_resources()

    def get_resource(self, resource_id):
        return self.resource_set.get_resource(resource_id)

    def get_resources_from_task_type(self, task):
        """add all the resources with type abilities in task.set_resources"""
        res_set = self.get_resources()
        for res in res_set:
            res = self.get_resource(res)
            for i in range(len(res.id_role)):
                if res.id_role[i] == task.task_type:
                    task.set_resources.append(res.id)
      
    # tasks methods ###########################################################

    def get_task(self, task_id):
        """return the task asociated to id"""
        return self.root_task.get_node_by_id(task_id)

    def get_nb_tasks(self):
        """returns number of task/milestone in project plus one
        (project itself)"""
        return len(self.root_task.flatten())

    def is_in_allocated_time(self, task_id, date):
        """
        tests if day is in allocated time for task t_id
        """
        for begin, end, _, _, _, _ in self.activities.select('task', task_id):
            if begin <= date < end:
                return True
        return False

    def set_task_property(self, task_id, attr, value):
        assert self._is_scheduled, "Project not scheduled"
        self.tasks.set_cell_by_ids(task_id, attr, value)

    def get_task_status(self, task):
        """
        return the status computed during the scheduling
        if task_id is not oredered, return default value todo
        """
        assert self._is_scheduled, "Project not scheduled"
        ordered_status = ['problem', 'todo', 'current', 'done']
        status_list = []
        for leaf in task.leaves():
            if leaf.TYPE == 'Milestone':
                # as leaf.id is int, self.tasks[leaf.id] will mistake it
                _, _, status, _, _ = self.tasks.get_row_by_id(leaf.id)
                status_list.append(status)
            for status in ordered_status:
                if status in status_list:
                    return status
        return 'todo'

    def task_has_info(self, task_id):
        """
        indicates if the task_id has any info stored in the object
        useful for example when writing xml for object
        """
        assert self._is_scheduled, "Project not scheduled"
        return len(self.tasks.get_row_by_id(task_id)) > 0
        return False

    def get_task_date_range(self, task):
        if task.TYPE == "milestone":
            date = self.milestones.get(task.id)
            return date, date

        begins, ends = [], []
        for leaf in task.leaves():
            for begin, end, __resource, __task, __usage, __src \
                    in self.activities.select('task', leaf.id):
                begins.append(begin)
                ends.append(end)
        try:
            begin = min(begins)
        except ValueError:
            raise ValueError('Task %s has no begin' % (task.id,))
        try:
            end = max(ends)
        except ValueError:
            raise ValueError('Task %s has no end' % task.id)
        return begin, end

    # activities methods ######################################################
    
    def add_activity(self, begin, end, resource_id, task_id,
                     usage=1., src='past'):
        """add an activity"""
        assert src in ['past', 'plan']
        self.activities.append_row([begin, end, resource_id, task_id,
                                    usage, src])

    def add_activities(self, rows):
        """concatenates another activities entity"""
        for row in rows:
            self.add_activity(*row)
        self.update_caches()
           
    def has_activity(self, t_id):
        return len(self.activities.select('task', t_id)) > 0

    def get_nb_of_activities(self):
        rowsize, colsize = self.activities.get_dimensions()
        return rowsize

    # usage & costs methods ###################################################

    def get_resources_duration_per_task(self, task_id):
        result = {}
        grouped = self.activities.groupby('task', 'resource')
        try:
            grouped = grouped[task_id]
        except KeyError:
            return {}
        for r_id, rows in grouped.items() :
            duration = 0.0
            for begin, end, resource, task, usage, src in rows:
                duration += self.compute_duration(begin, end, usage)
            result[r_id] = duration
        return result

    
    def get_usage(self, resource_id, task_id, date):
        """
        get the usage in percentage for the specified task, resource and date
        """
        grouped = self.activities.groupby('resource', 'task')
        # FIXME: remove try/catch?
        try:
            rows = grouped[resource_id][task_id]
        except KeyError:
            return 0.
        total_usage = 0.
        for begin, end, resource, task, usage, src in rows:
            if begin <= date <= end:
                total_usage += usage
        return total_usage

    def get_total_usage(self, resource_id, date):
        """
        get the usage in percentage at date for resource_id
        """
        total_usage = 0.0
        for begin, end, resource, task, usage, src \
                in self.activities.select('resource', resource_id):
            if begin <= date <= end:
                total_usage += usage
        return total_usage
    
    def get_resource_disponibility(self, resource_id, date):
        """get disponibility of a resource on a given date"""
        return 1 - self.get_total_usage(resource_id, date)
    
    def get_duration_worked_for_t(self, task_id):
        """gets the duration of work in days (float) of a given task"""
        warn('Project.get_duration_worked_for_t() is to be removed soon',
             DeprecationWarning)
        duration = 0.0
        for begin, end, resource, task, usage, src \
                in self.activities.select('task', task_id):
            duration += self.compute_duration(begin, end, usage)
        return duration

    
    def get_task_total_cost(self, task_id ,task_tot_duration):
        """
        obtain concatenation of costs for a task
        """
        costs = self.get_task_costs(task_id, task_tot_duration)[0]
        return sum(costs.values())

    def get_task_costs(self, task_id, task_tot_duration):
        """
        run through all activities and sum up duration * usage
        # FIXME - do we really need to send back durations
        from here? (simpler in tasks_costs_writer)
        """
        costs = {}
        durations = {}
        rounded=0
        if 0 < task_tot_duration % (1./self.factor) < (1./self.factor):
            rounded = task_tot_duration % (1./self.factor) - 1./self.factor
        for begin, end, resource, task, usage, src \
                in self.activities.select('task', task_id):
            costs.setdefault(resource, 0)
            durations.setdefault(resource, 0)
            # FIXME - presuming 8 hour /day work
            # FIXME - leaving behind the currency
            duration = self.compute_duration(begin, end, usage)
            tot_res_duration = 0
            durations[resource] += duration
            # calcul du total du temps consomme sur toutes les resources pour cette tache
            for res in durations:
                tot_res_duration += durations[res]
        # gestion des arrondis
        if len(durations) > 0:
            rounded  = rounded / len(durations)
            for res in durations:
                # using resources old definition
                task =  self.get_task(task_id)
                if self.resource_role_set.width() == 1 and not task.task_type:
                    try:
                        cost_rate = self.get_resource(res).hourly_rate[0]
                    except NodeNotFound,ex :
                        # if resource not found ???
                        cost_rate = 1
                #using new resources definition
                else:
                    role = task.task_type
                    cost_rate = self.get_cost_from_role(role)
                durations[res] += rounded
                costs[res] += durations[res] * cost_rate * HOURS_PER_DAY
        return costs, durations

    def get_cost_from_role(self, role):
        res_role = self.resource_role_set.get_resource_role(role)
        return res_role.hourly_cost

    def compute_duration(begin, end, usage):
        delta = end.day - begin.day
        days = 1 + delta
        return days * usage
    compute_duration = staticmethod(compute_duration)




#Copyright (c) 2000-2004 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Projman - (c)2004 Logilab - All rights reserved."""

__revision__ = "$Id: constants.py,v 1.3 2005-06-29 23:31:21 nico Exp $"

# needed for calculation with floats
FLOAT_ZERO = 0.000001
HOURS_PER_DAY = 8.
# task constraints
BEGIN_AFTER_END = 'begin-after-end'
BEGIN_AFTER_BEGIN = 'begin-after-begin'
END_AFTER_END = 'end-after-end'
END_AFTER_BEGIN = 'end-after-begin'
TASK_CONSTRAINTS = [BEGIN_AFTER_END, BEGIN_AFTER_BEGIN,
                    END_AFTER_END, END_AFTER_BEGIN]

# date constraints
BEGIN_AFTER_DATE = 'begin-after-date'
BEGIN_AT_DATE = 'begin-at-date'
BEGIN_BEFORE_DATE = 'begin-before-date'
END_AFTER_DATE = 'end-after-date'
END_AT_DATE = 'end-at-date'
END_BEFORE_DATE = 'end-before-date'
AT_DATE = 'at-date'
DATE_CONSTRAINTS = [BEGIN_AFTER_DATE, BEGIN_AT_DATE, BEGIN_BEFORE_DATE,
                    END_AFTER_DATE, END_AT_DATE, END_BEFORE_DATE]

__all__ = ['FLOAT_ZERO', 'BEGIN_AFTER_END', 'BEGIN_AFTER_BEGIN',
           'END_AFTER_END', 'END_AFTER_BEGIN', 'TASK_CONSTRAINTS',
           'BEGIN_AFTER_DATE', 'BEGIN_AT_DATE', 'BEGIN_BEFORE_DATE',
           'END_AFTER_DATE', 'END_AT_DATE', 'END_BEFORE_DATE',
           'DATE_CONSTRAINTS', ]


try:
    from projman.scheduling.gcsp import load_types
    TASK_SHARED = load_types.TASK_SHARED
    TASK_ONEOF = load_types.TASK_ONEOF
    TASK_SAMEFORALL = load_types.TASK_SAMEFORALL
    TASK_SPREAD = load_types.TASK_SPREAD
    TASK_MILESTONE = load_types.TASK_MILESTONE
except ImportError:
    TASK_SHARED = 0
    TASK_ONEOF = 1
    TASK_SAMEFORALL = 2
    TASK_SPREAD = 3
    TASK_MILESTONE = 4
    # for now just stop since we don't handle task types out of gcsp
    #raise

LOAD_TYPE_MAP = {
    "shared" : TASK_SHARED,
    "oneof" : TASK_ONEOF,
    "sameforall" : TASK_SAMEFORALL,
    "spread" : TASK_SPREAD,
    "milestone" : TASK_MILESTONE,
}
REVERSE_LOAD_TYPE_MAP = {}
for k,v in LOAD_TYPE_MAP.items():
    REVERSE_LOAD_TYPE_MAP[v] = k



# Copyright (c) 2000-2002 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" xml exportation utilities """

__revision__ = "$Id: projman_writer.py,v 1.17 2005-11-11 15:53:21 nico Exp $"

import logging
from logilab.common.visitor import Visitor
from projman.lib.constants import BEGIN_AT_DATE, END_AT_DATE, AT_DATE
from projman.lib.constants import REVERSE_LOAD_TYPE_MAP
from projman.lib.task import MileStone
try:
    import xml.etree.ElementTree as ET
except ImportError:
    import elementtree.ElementTree as ET
log = logging.getLogger("writer") # in case we use it one day


EXT_DATE_FORMAT = u'%Y-%m-%d'

#TEMP
import sys

# this function has been taken from http://infix.se/2007/02/06/gentlemen-indent-your-xml
def indent(elem, level=0):
    i = "\n" + level*"  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        for e in elem:
            indent(e, level+1)
            if not e.tail or not e.tail.strip():
                e.tail = i + "  "
        if not e.tail or not e.tail.strip():
            e.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i

# xml exportation methods ######################################################

def write_schedule_as_xml(filename, project):
    root = schedule_as_dom(project)
    indent(root)
    tree = ET.ElementTree(root)
    tree.write(filename, encoding="UTF-8")

def write_tasks_as_xml(filename, project):
    root = tasks_as_dom(project)
    indent(root)
    tree = ET.ElementTree(root)
    tree.write(filename, encoding="UTF-8")

# dom utilities ################################################################

def schedule_as_dom(project):
    """
    returns dom representation of project's schedule
    """
    doc = ET.Element("schedule")
    for task in project.root_task.leaves():
        if isinstance(task, MileStone):
            element = ET.SubElement( doc, 'milestone', id=task.id )
        else:
            element = ET.SubElement( doc, 'task', id=task.id )

        # add date-constraints
        begin, end = project.get_task_date_range(task)
        if element.tag == "task":
            ET.SubElement( element, "constraint-date", type=BEGIN_AT_DATE ).text = str(begin)
            ET.SubElement( element, "constraint-date", type=END_AT_DATE ).text = str(end)
            if task.priority is not None:
                ET.SubElement(element, 'priority').text = str(task.priority)
            ET.SubElement(element, 'status').text = project.get_task_status(task)
        else:
            ET.SubElement( element, "constraint-date", type=AT_DATE ).text = str(begin)

        # task_constraints
        for ctype, ctask_id, priority in task.task_constraints:
            ET.SubElement(element, "constraint-task", type=ctype, idref=ctask_id)

        if element.tag == "milestone":
            continue # milestone don't need more
        # global cost
        costs = project.get_task_costs(task.id, task.duration)[0]
        global_cost = ET.SubElement(element, 'global-cost', unit='XXX')
        global_cost.text = '%.1f' % sum(costs.values())
        # cost by resource
        if costs:
            costs_list = ET.SubElement(element, 'costs_list')
            for res_id, res_cost in costs.items():
                cost = ET.SubElement(costs_list, 'cost', idref=res_id)
                cost.text = '%.1f' % res_cost
        # report planned_activities only
        # (since only results of shceduling are written
        grouped = project.activities.groupby('src', 'task', 'resource')
        try:
            planned = grouped['plan'][task.id]
        except KeyError:
            continue

        rlist = ET.SubElement(element, 'report-list')
        for r_id, rows in planned.iteritems():
            for begin, end, resource, planned_task, usage, _ in rows:
                usage = usage or 1
                act_element = ET.SubElement(rlist, 'report',
                                            idref=r_id,
                                            usage='%f' % usage)
                act_element.set('from', str(begin))
                act_element.set('to', str(end) )
    return doc

class TasksVisitor:
    def __init__(self):
        self.parents = []

    def visit_root(self, node):
        elem = ET.Element('task', id=node.id)
        self.set_common_attr( node, elem )
        for rtype, rid in node.resource_constraints:
            ET.SubElement( elem, "constraint-resource",
                           idref=rid, type=rtype )
        
        self.parents.append(elem)
        for c in node.children:
            c.accept( self )
        return elem

    def visit_task(self, node):
        elem = ET.SubElement( self.parents[-1], 'task', id=node.id )
        elem.set("load-type", REVERSE_LOAD_TYPE_MAP[node.load_type] )
        elem.set("load", str(node.duration) )
        self.set_common_attr( node, elem )

        for rtype, rid in node.resource_constraints:
            ET.SubElement( elem, "constraint-resource",
                           idref=rid, type=rtype )
        
        self.parents.append( elem )
        for c in node.children:
            c.accept( self )
        self.parents.pop()

    def visit_milestone(self, node):
        elem = ET.SubElement( self.parents[-1], 'milestone', id=node.id )
        self.set_common_attr( node, elem )

    def set_common_attr(self, node, elem):
        if node.title:
            el = ET.SubElement( elem, "label")
            el.text = node.title
        if node.description_raw:
            if node.description_format=='docbook':
                text = u"<description format='docbook'>%s</description>" % node.description_raw
                el = ET.fromstring(text.encode("utf-8"))
                elem.append( el )
            elif node.description_format=='rest':
                el = ET.SubElement( elem, "description", format="rest" )
                el.text = node.description_raw
            else:
                text = u"<description>%s</description>" % node.description_raw
                el = ET.fromstring(text.encode("utf-8"))
                elem.append( el )
        for ctype, tid, priority in node.task_constraints:
            ET.SubElement( elem, "constraint-task", type=ctype, idref=tid )

        for ctype, date, priority in node.date_constraints:
            el = ET.SubElement( elem, "constraint-date", type=ctype )
            el.text = date.strftime("%F")

def tasks_as_dom(project):
    """
    returns dom representation of project's schedule
    """
    visitor = TasksVisitor()
    return visitor.visit_root( project.root_task )



##
# mini demo utilisation matplotlib pour du svg:
#
from math import sqrt
from matplotlib.backends.backend_svg import RendererSVG
from matplotlib.font_manager import FontProperties
import codecs
from StringIO import StringIO
from xml.sax.saxutils import escape

# matplotlib bugfix
from matplotlib import rcParams
for fam in ('serif', 'sans-serif', 'monospace', 'cursive', 'fantasy'):
    param = rcParams['font.'+fam]
    if isinstance(param,str):
        rcParams['font.'+fam] = param.split(', ')

# More matplotlib hacking : monkey patch draw_text so it uses a correct font spec
from matplotlib.backends.backend_svg import rgb2hex
def draw_text(self, gc, x, y, s, prop, angle, ismath):
    if ismath:
        self._draw_mathtext(gc, x, y, s, prop, angle)
        return

    #thetext = '%s' % s
    thetext = escape(s)
    fontfamily=prop.get_family()[0]
    fontstyle=prop.get_style()
    fontsize = prop.get_size_in_points()
    fontweight = prop.get_weight()
    color = rgb2hex(gc.get_rgb())

    style = 'font-size: %f; font-family: %s; font-style: %s; font-weight: %s; fill: %s;'%(fontsize, fontfamily,fontstyle,
                                                                                          fontweight, color)
    if angle!=0:
        transform = 'transform="translate(%f,%f) rotate(%1.1f) translate(%f,%f)"' % (x,y,-angle,-x,-y) # Inkscape doesn't support rotate(angle x y)
    else: transform = ''

    svg = """\
<text style="%(style)s" x="%(x)f" y="%(y)f" %(transform)s>%(thetext)s</text>
""" % locals()
    self._svgwriter.write (svg)

RendererSVG.draw_text = draw_text



def test():
    svgwriter = codecs.open( "test.svg", "w", "utf-8")
    renderer = RendererSVG( 800, 600, svgwriter, "test" )
    gc = renderer.new_gc()
    gc.set_linewidth(4)
    renderer.draw_line(gc, 0,0,800,600)
    renderer.draw_line(gc, 0,600,800,0)
    renderer.finish()
    svgwriter.close()

from pil import COLORS, delta_color as _delta_color

def _color( icol ):
    """Convert (r,g,b) with range 0..255 to
    rgb range 0..1"""
    if icol in COLORS:
        icol = COLORS[icol]
    if isinstance(icol, (int,float)):
        icol = (icol,icol,icol)
    return tuple([ c/255. for c in icol])

class SVGHandler:
    def __init__(self, format):
        pass

    def get_gc( self,
                color=None,
                fillcolor=(0,0,0),
                delta_color=None,
                **args ):
        gc = self._rend.new_gc()
        col = color or fillcolor
        if not col and delta_color:
            col = _delta_color(delta_color, int(value)) # FIXME value ??
        gc.set_foreground( _color( col ) )
        return gc

    def get_prop(self,
                 style=None,
                 weight=None,
                 **args ):
        prop = FontProperties()
        prop.set_family('sans-serif')
        prop.set_size(10)
        if style:
            prop.set_style(style)
        if weight:
            prop.set_weight(weight)
        return prop

    def init_drawing(self, width, height):
        """ initialize a new picture """
        self._buffer = StringIO()
        self._rend = RendererSVG( width, height, self._buffer, "projman" )
        self.height=height

    def close_drawing(self, cropbox=None):
        """ close the current picture """
        pass

    def get_result(self):
        """ return the current picture """
        self._rend.finish()
        return self._buffer.getvalue()

    def save_result(self, stream):
        """ return the current picture """
        self._rend.finish()
        data = self._buffer.getvalue()
        stream.write( data )

    def open_link(self,url):
        """draw a link to an url"""
        self._rend._svgwriter.write('<a xlink:href=\"%s\"> ' %url)

    def close_link(self):
        self._rend._svgwriter.write('</a>')

    def draw_text(self, x, y, text, **args):
        """ draw a text """
        #print "draw_text", x, y, repr(text), args
        n=len(text)
        text=text.lstrip()
        gc = self.get_gc( **args )
        prop = self.get_prop( **args )
        if n!=len(text): # spaces are not printed so we shift left n*size(i)
            try:
                w,h,_ = self._rend.get_text_width_height_descent("i", prop, False)
            except AttributeError: # compatibility with matplotlib 0.90
                w,h = self._rend.get_text_width_height("i", prop, False)
            x+= w*(n-len(text))
        self._rend.draw_text( gc, x, y,
                              text, prop,
                              angle=0, ismath=False)
        
    def draw_line(self, x1, y1, x2, y2, **args):
        """ draw a line """
        y1 = self.height-y1
        y2 = self.height-y2
        gc = self.get_gc( **args )
        gc._linewidth /= 2.
        self._rend.draw_line(gc, x1, y1, x2, y2)

    def draw_dot(self, x1, y1, x2, y2, n, **args):
        """ draw a dot line """
        dist = sqrt((x1-x2)**2 + (y1-y2)**2)
        length = dist / n
        dx = (x1 - x2)/length
        dy = (y1 - y2)/length
        gc = self.get_gc( **args )
        gc._linewidth /= 3.
        count=0
        while dist >= 0.00001:
            x1 -= 2*dx
            y1 -= 2*dy
            dist = sqrt((x1-x2)**2 + (y1-y2)**2)
            self.draw_line(x1, y1, x1+dx, y1+dy, **args)
            count+=1
            if count>1000:
                break

    def draw_rect(self, x, y, width, height, **args):
        """ draw a rectangle """
        #print "draw_rectangle", x, y, width, height, args
        y = self.height-y-height
        gc = self.get_gc( **args )
        rgbFace = _color( args['fillcolor'] )
        self._rend.draw_rectangle(gc, rgbFace, x, y, width, height)

    def draw_poly(self, point_list, **args):
        """ draw a polygon """
        #print "draw_poly", point_list, args
        gc = self.get_gc( **args )
        rgbFace = gc.get_rgb()
        h=self.height
        points = [ (x, h-y) for x,y in point_list ]
        self._rend.draw_polygon(gc, rgbFace, points)

    def get_output(self, fname):
        return  codecs.open(fname, "w", "utf-8")
